/*
  polymin.cpp
*/

#include <cstdlib>
#include <iostream>
#include <sdpsl.h>

using namespace std;
using namespace sdp;

//
// Prototypes
//
static void poly_min_sdp(const PolynomialVariable& x, const Polynomial& p);

/*
  main
*/
int main()
{
  PolynomialVariable x("x");
  Polynomial p = (x + 1) * (x + 2) * (x + 3) * (x - 2) * (x - 4) * (x - 5);
  
  poly_min_sdp(x, p);
  
  return 0;
}


/*
  poly_min_sdp

  Given a univariate polynomial p on variable x, outputs an SDP whose
  optimal value is the global minimum of p. Here we use the standard
  monomial basis.
*/
void poly_min_sdp(const PolynomialVariable& x, const Polynomial& p)
{
  /* We first find out the degree of the polynomial */
  int deg = 0;
  
  for (auto& c : p)
    if (c.first[x] > deg)
      deg = c.first[x];
    else if (c.first[x] < 0) {
      cerr << "Negative exponents are not supported." << endl;
      exit(-1);
    }
  
  if (deg % 2 != 0) {
    cerr << "Degree must be even!" << endl;
    exit(-1);
  }
  
  /* Then we build the problem */
  SDPVariable X(1 + deg / 2);
  SDPVariable s(1), t(1);

  /* First we construct a moment matrix to express the polynomial p as
   * a sum of squares. 
   */
  Matrix<Polynomial> R(1 + deg / 2);
  
  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++)
      R[i][j] = x(i + j);
  
  /* Then we create the constraints */
  SDPConstraintContainer cc;
  
  cc.add_constraint(R * X + Matrix<Polynomial>(1, 1, 1) * s
		    + Matrix<Polynomial>(1, 1, -1) * t == p);
  
  /* And the problem with its objective function */
  SDPProblem problem;

  problem.set_objective(Matrix<double>(1, 1, 1) * s + Matrix<double>(1, 1, -1) * t);
  problem << cc;
    
  /* And we write out the SDPA file */
  cout << SDPAWriter(problem);
}
