/*
  polymin.cpp

  Outputs, in sparse SDPA format, an SDP whose optimal value is the
  global minimum of a univariate polynomial. We use a basis of
  Laguerre polynomials to represent the polynomial constraint.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdlib>
#include <iostream>
#include <sdpsl.h>

using namespace std;
using namespace sdp;

//
// Prototypes
//
static void poly_min_sdp(const PolynomialVariable& x, const Polynomial& p);

/*
  main
*/
int main()
{
  PolynomialVariable x("x");
  Polynomial p = (x + 1) * (x + 2) * (x + 3) * (x - 2) * (x - 4) * (x - 5);
  
  poly_min_sdp(x, p);
  
  return 0;
}


/*
  poly_min_sdp

  Given a univariate polynomial p on variable x, outputs an SDP whose
  optimal value is the global minimum of p. Here we use a basis of
  Laguerre polynomials.
*/
void poly_min_sdp(const PolynomialVariable& x, const Polynomial& p)
{
  /* We first find out the degree of the polynomial */
  int deg = 0;
  
  for (auto& c : p)
    if (c.first[x] > deg)
      deg = c.first[x];
    else if (c.first[x] < 0) {
      cerr << "Negative exponents are not supported." << endl;
      exit(-1);
    }
  
  if (deg % 2 != 0) {
    cerr << "Degree must be even!" << endl;
    exit(-1);
  }
  
  /* Then we build the problem */
  SDPVariable X(1 + deg / 2);
  SDPVariable s(1), t(1);

  /* First we construct a moment matrix to express the polynomial p as
   * a sum of squares. 
   */
  Matrix<Polynomial> R(1 + deg / 2);
  
  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++)
      R[i][j] = x(i + j);

  /* We create a basis of Laguerre polynomials */
  vector<Polynomial> laguerre_basis;
  compute_laguerre_polys(deg, x, 0, laguerre_basis);

  /* And we use it to add the polynomial constraint */
  SDPPolynomialConstraint C = R * X + Matrix<Polynomial>(1, 1, 1) * s
    + Matrix<Polynomial>(1, 1, -1) * t == p;

  SparseBasisChanger sc(SparseMonomialBasis(C), laguerre_basis);

  SDPConstraintContainer cc;
  cc.add_constraint(C, sc);
  
  /* And the problem with its objective function */
  SDPProblem problem;

  problem.set_objective(Matrix<double>(1, 1, 1) * s + Matrix<double>(1, 1, -1) * t);
  problem << cc;
    
  /* And we write out the SDPA file */
  cout << SDPAWriter(problem);
}
