/*!
  \file csdpreader.h
  
  \brief Reads a solution from an SDP given in CSDP solution format.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __CSDPREADER_H__
#define __CSDPREADER_H__

#include <list>
#include <iostream>
#include <string>
#include <utility>
#include <cctype>
#include <cstdlib>
#include <climits>
#include <initializer_list>
#include <sdpsl/sdpenv.h>
#include <sdpsl/sdperror.h>
#include <sdpsl/sdpproblem.h>
#include <sdpsl/sdpsolution.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Reads a solution from an SDP given in CSDP solution format.

  Environment parameter \c output_eps is used to decide which numbers
  are treated as nonzeros. Only numbers whose absolute value is
  greater than \c output_eps are considered nonzeros.

  The reader can read solutions in double-precision or
  arbitrary-precision, depending on whether \c T is \c double or \c
  mpf_class.

  \sa \c SDPEnvironment
*/
template <class T>
class CSDPReader {
  SDPEnvironment env;
  std::list<SDPVariable> variables;
  SDPSolution<T>& solution;

  bool str_to_int(const char *str, int *result) const;

  void read_solution(std::istream& input, 
		     const std::list<SDPVariable>& ignore_list = std::list<SDPVariable>());

public:
  /*!
    \brief Creates a reader given a problem and a solution map.

    The reader is created given a problem and a solution map, an
    instance of \c SDPSolution that will hold the solution. The
    problem is used only so that the reader knows which variables are
    used.

    \warning This class keeps a reference to the solution map
    given. It is the caller's job to guarantee that the solution map
    lives through the existence of the reader.
  */
  CSDPReader(const SDPProblem& problem, SDPSolution<T>& sol, 
             const SDPEnvironment& e = SDPEnvironment())
    : env(e), variables(problem.variables()), solution(sol) { }

  /*!
    \brief Creates a reader given a list of variables and a solution map.

    The reader is created given a list of variables and a solution
    map. This can be used when one wishes to avoid building the
    problem.

    \warning This class keeps a reference to the solution map
    given. It is the caller's job to guarantee that the solution map
    lives through the existence of the reader.
  */
  CSDPReader(const std::list<SDPVariable> var_list, SDPSolution<T>& sol, 
             const SDPEnvironment& e = SDPEnvironment())
    : env(e), variables(var_list), solution(sol) { }

  /*!
    \brief Creates a reader given a list of variables and a solution map.

    The reader is created given a list of variables and a solution
    map. This can be used when one wishes to avoid building the
    problem, what usually includes the time-consuming step of creating
    constraints.

    \warning This class keeps a reference to the solution map
    given. It is the caller's job to guarantee that the solution map
    lives through the existence of the reader.
  */
  CSDPReader(const std::initializer_list<SDPVariable> var_list, SDPSolution<T>& sol, 
             const SDPEnvironment& e = SDPEnvironment())
    : env(e), variables(var_list), solution(sol) { }

  /*! 
    \brief Reads the solution from a given stream.

    \throws ReadError if there is a problem with the input file. An
    error message is stored and can be retrieved with method
    <tt>what</tt> from the exception and the number of the line where
    the error occurred can be retrieved with the method
    <tt>lineno</tt>.
  */
  friend std::istream& operator>>(std::istream& input, CSDPReader reader)
  {
    reader.read_solution(input);
    return input;
  }
};

/*
  str_to_int

  Converts string to integer, returns true if successful, false
  otherwise.
*/
template <class T>
bool CSDPReader<T>::str_to_int(const char *str, int *result) const
{
  char *endptr;
  long ret = strtol(str, &endptr, 10);

  if (*endptr) return false;
  if (ret > INT_MAX) return false;

  *result = (int) ret;

  return true;
}

/*
  read_solution

  ignore_list is a list of variables that should be ignored in the
  output; it is not anymore used, should be removed in the future.
*/
template <class T>
void CSDPReader<T>::read_solution(std::istream& input, const std::list<SDPVariable>& ignore_list)
{
  /* We compute the block numbers corresponding to variables, given
     the ignore_list. Block numbers are already +1.
  */
  std::vector<int> block_number(SDPVariable::num_variables(), -1);

  for (auto& var : variables)
    block_number[var.idx()] = 0;  // These are the variables actually used

  for (auto& var : ignore_list)
    block_number[var.idx()] = -1;  // Marks that this variable is ignored

  int num_blocks = 0;

  for (int i = 0; i < block_number.size(); i++)
    if (block_number[i] == 0)
      block_number[i] = ++num_blocks;
  
  /* We first make a vector containing for every block its associated
     matrix and mask for output, if one is given. We take the
     opportunity to clear the masks (when given) and the matrices,
     even for the ignored variables.
     
     Recall block numbers start at 1.
  */
  typedef std::pair<Matrix<T> *, MatrixMask *> value_type;

  std::vector<value_type> blocks(num_blocks + 1, value_type(nullptr, nullptr));

  for (auto& c : solution) {
    if (block_number[c.first.idx()] != -1)
      blocks[block_number[c.first.idx()]] = value_type(&c.second.matrix(), c.second.mask());
    
    if (c.second.mask())
      c.second.mask()->clear();
    else {
      Matrix<T>& A = c.second.matrix();

      for (int i = 0; i < A.num_rows(); i++)
        for (int j = 0; j < A.num_cols(); j++)
          A[i][j] = 0;
    }
  }

  /* Then we start reading the data */
  std::string line;
  int lineno = 0;
  double output_eps = env.output_eps();
  const char *words[10];

  while (!input.eof()) {
    lineno++;

    if (!getline(input, line)) break;

    /* We care only for lines starting with a "2 " */
    if (line.size() < 2 || line[0] != '2' || line[1] != ' ')
      continue;

    /* We first split the string at the spaces */
    for (int i = 0; i < line.size(); i++)
      if (isspace(line[i])) line[i] = '\0';

    int num_words = 0;
    const char *c_str = line.c_str();
    for (int i = 0; i < line.size(); ) {
      if (c_str[i]) {
	if (num_words == 5)
	  throw ReadError(ERROR_MSG("Error reading from stream"), lineno);

	words[num_words++] = c_str + i;

	while (c_str[i]) i++;
      }
      else i++;
    }

    if (num_words < 5)
      throw ReadError(ERROR_MSG("Error reading from stream"), lineno);

    /* Then we read the info */
    int b, i, j;
    
    if (!str_to_int(words[1], &b) || !str_to_int(words[2], &i) 
	|| !str_to_int(words[3], &j))
      throw ReadError(ERROR_MSG("Error reading from stream"), lineno);

    i--;
    j--;

    assert(b >= 1);

    if (b > num_blocks) continue;  // The slack block must be ignored!

    if (blocks[b].first) {
      T num;
      try {
        NumberOperations::from_string(words[4], num);
      }
      catch (ValueError) {
        throw ReadError(ERROR_MSG("Error converting number"), lineno);
      }

      /* Sets the matrix entries */
      Matrix<T>& A = *blocks[b].first;
      
      if (!(i >= 0 && i < A.num_rows())
          || !(j >= 0 && j < A.num_cols()))
        throw ReadError(ERROR_MSG("Invalid matrix position"), lineno);

      if (NumberOperations::abs(num) <= output_eps) 
        continue;

      A[i][j] = A[j][i] = num;

      if (blocks[b].second) {
        blocks[b].second->turn_on(i, j);
	blocks[b].second->turn_on(j, i);
      }
    }
  }
}


}  // namespace

#endif

// Local variables:
// mode: c++
// End:
