/*!
  \file matrix.h
  
  \brief Class for matrices.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MATRIX_H__
#define __MATRIX_H__

#include <vector>
#include <memory>
#include <iostream>
#include <initializer_list>
#include <sdpsl/sref.h>
#include <sdpsl/sdperror.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// Class declaration
//

/*!
  \brief A matrix whose elements are of type \c T.

  This template class represents a matrix whose elements are of type \c
  T. The matrix can be square or not. It can be accessed using the
  usual <tt>C</tt> notation.

  \remark Matrices are stored in row-major order, and their data
  vectors (the vector where the matrix is stored, one row after the
  other) can also be accessed.

  \section ref References to matrices

  By default, matrices behave as any other class. They have copy
  constructors that behave in the standard way, so that when a matrix
  is initialized from another matrix, data is copied.

  There is also a mechanism to make copies by reference from a
  matrix. When a copy by reference is made, a new matrix is created
  containing a pointer to the data of the original matrix, so changes
  to it mean changes to the original matrix. Since the data is not
  copied, creating a reference is a constant-time operation.

  Unlike a \cpp reference however, such <i>shared references</i>
  are still valid after the original matrix is destroyed. When this
  happens, the shared reference takes over the responsibility to free
  the memory occupied by the matrix data. More than one shared
  reference can be taken from the same matrix, and only when the last
  one is destroyed will the data of the original matrix be destroyed
  (this is implemented via \c std::shared_ptr).

  Here is an example of how to take and use shared references:

  \include matref.cpp

  \sa \c shared_ref, \c const_shared_ref
  
*/
template <class T>
class Matrix {
  /* Friends */
  friend class shared_ref<Matrix<T> >;
  friend class const_shared_ref<Matrix<T> >;

  /* Private types */
  typedef std::shared_ptr<std::vector<T> > data_ptr;

  /* Data */
  int m, n;  // Number of rows and columns
  data_ptr data;

  /* Mechanism for taking references */
  Matrix(const Matrix& other, const data_ptr& _data)
    : m(other.m), n(other.n), data(_data) {  }

  const data_ptr& take_data() const { return data; }

public:
  //! Copy constructor.
  Matrix(const Matrix<T>& other);

  explicit Matrix(int mm, int nn = -1, const T& val = T());
  explicit Matrix(const std::initializer_list<std::initializer_list<T> >& init_data);

  //! Virtual destructor.
  virtual ~Matrix() { }

  /* Conversion constructor */
  template <class U, class Converter>
  explicit Matrix(const Matrix<U>& other, const Converter& converter);

  /* Get and set */
  int num_rows() const { return m; }  //!< Returns number of rows of matrix.
  int num_cols() const { return n; }  //!< Returns number of columns of matrix.

  bool is_square() const { return m == n; }  //!< Returns true if matrix is square.

  /*!
    \brief Returns a pointer to the vector containing the data of the matrix.

    Matrices are stored in row-major order, that is, their rows are
    stored, one after the other, in a vector. This function returns
    the vector containing the rows of the matrix, one after the other.
  */
  T *vector_data() { return data->data(); }

  //! \c const version of \c vector_data.
  const T *vector_data() const { return data->data(); }

  //! Returns the size of the underlying vector that contains the data.
  int vector_size() const { return m * n; }

  /* Copy assignment */
  Matrix<T>& operator=(const Matrix<T>&);

  /* Indexing */

  /*!
    \brief Returns a pointer to the first element of row \a row.

    Matrices are stored in row-major order. This function returns a
    pointer to the first element of the row with index \a row of the
    matrix. This means that the element in row \a i and column \a j of
    matrix \a A can therefore be accessed as <tt>A[i][j]</tt>.
  */
  T *operator[](int row) { return data->data() + row * n; }

  //! \c const version of \c operator[].
  const T *operator[](int row) const { return data->data() + row * n; }

  /* Operators */
  Matrix<T>& operator*=(const T& val);

  /*!
    \brief Returns the result of multiplying every element of \a A by an instance of \a val.

    This function assumes objects of type \c T can be multiplied. If
    this is not the case, there will be no compile-time error, unless
    this operator is used and therefore this code is instantiated.
  */
  friend Matrix<T> operator*(const T& val, const Matrix<T>& A)
  {
    Matrix<T> ret = A;
    ret *= val;
    return ret;
  }

  /* Static functions */
  static Matrix<T> identity(int n);

  /*!
    \brief Outputs the matrix to a stream, in SAGE-friendly format.
  */
  friend std::ostream& operator<<(std::ostream& out, const Matrix<T>& A)
  {
    out << "[ ";
    for (int i = 0; i < A.num_rows(); i++) {
      out << "[ ";

      for (int j = 0; j < A.num_cols(); j++) {
        if (j > 0) out << ", ";
	out << A[i][j];
      }

      out << " ]";
      if (i < A.num_rows() - 1)
        out << ", " << std::endl;
    }

    out << " ]" << std::endl;

    return out;
  }
};

//
// Method definition
//

/*
  Copy constructor
*/
template <class T>
Matrix<T>::Matrix(const Matrix<T>& other)
  : m(other.m), n(other.n), 
    data(std::make_shared<std::vector<T> >(*other.data))
{

}

/*!
  \brief Initializes a matrix of elements of type \c T from a matrix
  of elements of type \c U.

  Given a matrix with elements of type \c U, converts it to a matrix
  with elements of type \c T. To make the conversion a \e converter
  class is used. This is any class which implements the function call
  operator. If \c converter is such a class and if \c val is of type
  \c U, then <tt>converter(val)</tt> must be of type \c T.

  \include matconv.cpp
*/
template <class T>
template <class U, class Converter>
Matrix<T>::Matrix(const Matrix<U>& other, const Converter& converter)
  : m(other.num_rows()), n(other.num_cols()),
    data(std::make_shared<std::vector<T> >(m * n))
{
  for (int i = 0; i < m * n; i++)
    data->at(i) = converter(other.vector_data()[i]);
}

/*!
  \brief Constructs a matrix with \a mm rows and \a nn columns, with
  elements initialized to \a val.
*/
template <class T>
Matrix<T>::Matrix(int mm, int nn, const T& val)
{
  m = mm;

  if (nn == -1) n = m;
  else n = nn;
  
  data = std::make_shared<std::vector<T> >(m * n, val);
}

/*!
  \brief Constructs a matrix from data given in an \c initializer_list.

  Given a list of lists, constructs a matrix based on it, each list
  giving a row of the matrix, as in the following example, where the matrix
  \f$\begin{pmatrix}1&2\\3&4\end{pmatrix}\f$ is created.

  \code
  Matrix<double> A({ { 1, 2 }, { 3, 4 } });
  \endcode

  \throws ValueError if the rows are not all of the same size or if
  the list of rows is empty.
*/
template <class T>
Matrix<T>::Matrix(const std::initializer_list<std::initializer_list<T> >& init_data)
{
  m = init_data.size();
  if (!m) throw ValueError(ERROR_MSG("Empty list of matrix rows"));

  n = init_data.begin()->size();

  data = std::make_shared<std::vector<T> >(m * n);

  int i = 0;
  for (auto& row : init_data) {
    if (row.size() != n) throw ValueError(ERROR_MSG("Rows of different sizes given"));
    
    int j = 0;
    for (auto& val : row) {
      (*data)[i * n + j] = val;

      j++;
    }

    i++;
  }
}


/*!
  \brief Assignment operator. Makes this matrix equal to \a other.
*/
template <class T>
Matrix<T>& Matrix<T>::operator=(const Matrix<T>& other)
{
  m = other.m;
  n = other.n;
  data = std::make_shared<std::vector<T> >(*other.data);
  
  return *this;
}

/*!
  \brief In-place multiplication by scalar.

  Multiplies every entry of the matrix by \a val. It is assumed the
  entries are of a type that supports multiplication. If not, then a
  compile-time error will occur.
*/
template <class T>
Matrix<T>& Matrix<T>::operator*=(const T& val)
{
  for (int i = 0; i < m * n; i++)
    data->at(i) *= val;

  return *this;
}

/*!
  \brief Returns an \a n by \a n identity matrix.
*/
template <class T>
Matrix<T> Matrix<T>::identity(int n)
{
  Matrix<T> ret(n, n, 0);

  for (int i = 0; i < n; i++)
    ret[i][i] = 1;

  return ret;
}

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
