/*!
  \file matrixmask.h

  \brief Binary mask for matrices

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MATRIXMASK_H__
#define __MATRIXMASK_H__

#include <memory>
#include <list>
#include <vector>
#include <sdpsl/sref.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Implements a matrix bit-mask.

  A \c MatrixMask contains an \a m by \a n matrix of bits. It can be
  used, for instance, to store the nonzero positions of a matrix. Some
  functions use a \c MatrixMask in order to efficiently deal with
  sparse matrices.

  A \c MatrixMask also contains a list of the entries that are \e on. So
  it is efficient to iterate over all the entries turned on.

  Each entry of the mask occupies only one bit, so that even large
  masks do not use up too much space. So it is efficient to use masks
  of large size. Notice however that, if too many positions are on,
  then the list of positions that are turned on might come to occupy a
  large amount of memory. So masks are better kept sparse.

  \c MatrixMask implements the <em>shared reference</em> mechanism. To
  understand how it works, see the documentation of \c shared_ref, \c
  const_shared_ref, and \c Matrix for an example.

  \par Complexity
  Turning a position on and checking whether a position is on or off
  takes constant time. Turning a position off takes at the moment time
  linear time on the number of positions that are on.  Iteration
  through the list of positions that are on takes constant time per
  position turned on.

  \section iter Iteration

  \c MatrixMask supports both \cpp and \cpp11 style
  iteration. One iterates over a list of \e positions, i.e.,
  instances of \c struct \c MatrixMask::Position, containing the
  nonzero entries of the mask.

  \todo In the future, it might be wise to change the list of
  positions \c on_pos to a vector. This might save memory.
*/
class MatrixMask {
public:
  //! A position in the mask
  struct Position {
    int i;  //!< Row of the entry
    int j;  //!< Column of the entry

    Position() : i(0), j(0) { }
    Position(int ii, int jj) : i(ii), j(jj) { }

    bool operator==(const Position& other) const { return i == other.i && j == other.j; }
  };

private:
  /* Shared refs stuff */
  friend class shared_ref<MatrixMask>;
  friend class const_shared_ref<MatrixMask>;

  struct Data {
    std::vector<char> mask;      // Mask data
    std::list<Position> on_pos;  // Nonzero positions

    explicit Data(int mask_size) : mask(mask_size, 0) { }
  };

  typedef std::shared_ptr<Data> data_ptr;

  /* Data */
  int m, n;       // Number of rows and columns
  data_ptr data;  // The data contained by the mask

  /* Mechanism for taking shared references */
  MatrixMask(const MatrixMask& other, const data_ptr& _data)
    : m(other.m), n(other.n), data(_data) { }

  const data_ptr& take_data() const { return data; }

public:
  //! Constructs an \a nn by \a nn mask with all entries off.
  explicit MatrixMask(int nn);

  //! Constructs an \a mm by \a nn mask with all entries off.
  MatrixMask(int mm, int nn);

  //! Copy constructor.
  MatrixMask(const MatrixMask& other);

  //! Assignment is not implemented.
  MatrixMask& operator=(const MatrixMask&) = delete;

  /* Get and set */
  int on_count() const { return data->on_pos.size(); }  //!< Returns number of on positions.
  int num_rows() const { return m; }  //!< Returns number of rows.
  int num_cols() const { return n; }  //!< Returns number of columns.

  //! Returns true if entry <tt>(i, j)</tt> of the mask is on.
  bool is_on(int i, int j) const;

  //! Turns on entry <tt>(i, j)</tt> of the mask.
  void turn_on(int i, int j);

  //! Turns off entry <tt>(i, j)</tt> of the mask.
  void turn_off(int i, int j);

  //! Clears the whole mask, turning off all entries.
  void clear();

  //! Constant iterator.
  typedef std::list<Position>::const_iterator const_iterator;

  //! Returns iterator to the beginning of the list of \e on positions.
  const_iterator begin() const { return data->on_pos.begin(); }

  //! Returns past-the-end iterator to the list of \e on positions.
  const_iterator end() const { return data->on_pos.end(); }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
