/*!
  \file polynomial.h

  \brief Class implementing a polynomial.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __POLYNOMIAL_H__
#define __POLYNOMIAL_H__

#include <map>
#include <iostream>
#include <sdpsl/number.h>
#include <sdpsl/monomial.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//! Class implementing a multivariate polynomial.
/*!
  A polynomial is a map assigning nonzero coefficients to monomials. It is a
  container whose elements are instances of \c std::pair<\c const \c Monomial,
  \c Number>.

  \remark The current implementation uses \c std::map, and this
  influences the complexity of operations as related below.

  \section create Creating polynomials

  Polynomials can be created using a syntax close to that of
  mathematical software such as SAGE. First, variables must be
  defined, and then used to define the polynomials in a simple
  way. Consider the example:

  \include polybasic.cpp

  \section iter Iteration 

  Polynomials can be iterated using \cpp11 notation or standard \cpp
  notation. Only constant iteration is allowed. Iteration is over the
  pairs contained in the polynomial. Pairs are ordered increasingly
  according to the lexicographical ordering on the monomials.

  Here is an example of iteration over a polynomial \c p.

  \code
  for (const auto& c : p)
    cout << "Coefficient of monomial " << p.first << " is " << p.second << endl;
  \endcode  

  \warning The use of operations that change the polynomial during
  iteration can invalidate iterators. No guarantees are given.
*/
class Polynomial {
  typedef std::map<Monomial, Number> MonomialMap;
  typedef MonomialMap::value_type value_type;

  /* Static variables */
  static const Number zero;

  /* Data */
  MonomialMap monomials;

  /* Utility */
  void add_monomial(const Number& coeff, const Monomial& mon);

public:
  //! Constructs a constant polynomial with constant term equal to \a val.
  Polynomial(const Number& val = 0);

  //! Constructs a polynomial consisting of one monomial \a x with coefficient 1.
  Polynomial(const PolynomialVariable& x);

  //! Constructs a polynomial consisting of a monomial equal to the given power \a pw with coefficient 1.
  Polynomial(const PolynomialVariable::Power& pw);
  
  //! Constructs a polynomial consisting of the given monomial \a mon with coefficient 1.
  Polynomial(const Monomial& mon);
  
  #ifdef __MPF_BUILD__
  Polynomial(double val);
  #endif

  /* Get and set */

  //! Returns the number of monomials in the polynomial.
  int size() const { return monomials.size(); }

  //! Returns the coefficient of a given monomial.
  const Number& coefficient(const Monomial& m) const
  {
    auto it = monomials.find(m);
    
    if (it == monomials.end()) return zero;
    else return it->second;
  }

  //! Constant iterator.
  typedef MonomialMap::const_iterator iterator;

  iterator begin() const { return monomials.begin(); }  //!< Returns iterator to beginning
  iterator end() const { return monomials.end(); }  //!< Returns iterator to the end

  /* Operators */

  //! In-place addition
  /*!
    Adds polynomial \a other to this polynomial, and returns this polynomial.
  */
  Polynomial& operator+=(const Polynomial& other);

  friend Polynomial operator+(const Polynomial& p, const Polynomial& q);

  /*!
    \brief In-place subtraction.

    Subtracts polynomial \a other from this polynomial and returns this
    polynomial.
  */
  Polynomial& operator-=(const Polynomial& other);

  friend Polynomial operator-(const Polynomial& p, const Polynomial& q);

  /*!
    \brief  In-place multiplication of polynomials.
    
    Multiplies this polynomial by a polynomial \a other and returns
    this polynomial.
  */
  Polynomial& operator*=(const Polynomial& other);

  friend Polynomial operator*(const Polynomial& p, const Polynomial& q);
  
  //! Returns the polynomial multiplied by -1.
  Polynomial operator-() const;

  /*!
    \brief In-place division by scalar.
    
    Divides every coefficient of the polynomial by the scalar \a val
    and returns the polynomial.
  */
  Polynomial& operator/=(const Number& val);

  friend Polynomial operator/(const Polynomial& p, const Number& val);

  friend std::ostream& operator<<(std::ostream& out, const Polynomial& p);

  /*!
    \brief Evaluates a univariate polynomial on variable \a x at \a val.

    Given a univariate polynomial on variable \a x and a number \a val,
    evaluates the polynomial at \a val and returns the result. We assume
    that all variables have nonnegative exponents.

    \throws PolyError::ValueError if it is detected that the
    polynomial is not univariate or if variable \a x occurs with a
    negative exponent.
 
    \warning The function assumes the polynomial being evaluated is
    univariate, but it does not explicitly test it. If the polynomial
    is not univariate, it might be that an exception will be thrown, or
    it might be that it will be evaluated as if it were a univariate
    polynomial on the specified variable (in which case all other
    variables are just considered to have exponent zero). This
    implementation choice aims at making the function more efficient.
  */
  Number univariate_eval(const PolynomialVariable& x, const Number& val) const;
};


//
// Function prototypes
//

//! Adds two polynomials and returns the result.
extern Polynomial operator+(const Polynomial& p, const Polynomial& q);

//! Subtracts two polynomials and returns the result.
extern Polynomial operator-(const Polynomial& p, const Polynomial& q);

//! Multiplies two polynomials and returns the result.
extern Polynomial operator*(const Polynomial& p, const Polynomial& q);

//! Divides polynomial \a p by scalar \a val and returns the result.
extern Polynomial operator/(const Polynomial& p, const Number& val);

/*!
  \brief Outputs polynomial to stream.
  
  Outputs polynomial \a p to stream \a out. For the output, monomials
  are ordered decreasingly according to the lexicographical ordering,
  so that monomials higher in the lexicographical ordering appear
  first.
*/
extern std::ostream& operator<<(std::ostream& out, const Polynomial& p);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
