/*!
  \file progbar.h

  \brief A character-based, thread-safe progress bar.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __PROGBAR_H__
#define __PROGBAR_H__

#include <mutex>
#include <string>
#include <iostream>
#include <sdpsl/sdpprogress.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A character-based, thread-safe progress bar.

  This class implements a simple character-based progress bar. It
  implements the \c SDPProgressIndicator interface and can be used
  with functions of \sdpsl that accept instances of \c
  SDPProgressIndicator.

  The progress bar shows a message and a number of characters to
  indicate progress, as well as a numerical counter. Here is a quick
  example:

  \include progress.cpp

*/
class ProgressBar : public SDPProgressIndicator {
  /* User supplied data */
  char _bar_char;             // Character used for bar
  std::string message_text;   // Message shown
  int _bar_size;              // Size of bar in characters
  int _max_value;             // Maximum value for progress
  int _update_step;           // Step for update, value of one char
  std::ostream& _out;         // Stream for output

  /* For internal use */
  int counter_wd;             // Number of digits in max_value
  double unit_value;          // Value of one character of the bar, except for the last one
  int cur, show_cur;          // Current value, current value being shown
  double cur_update_step;     // Current update step
  std::mutex update_mutex;    // For thread support during updates

  /* Private methods */
  void update();

public:
  /*!
    \brief Constructor for a progress bar with a given message and size.

    \param message is the message displayed.

    \param bar_size is the size of the progress bar in number of characters.

    \param max_value is the maximum value that the bar can attain.

    \param update_step is interval at which the bar is updated; if
    equal to zero then it is automatically set so that updates happen
    when one full-character should be added.

    \param bar_char is the character used to display progress.

    \param out is the output stream to which the bar is printed; default is \c cout.
  */
  ProgressBar(const std::string& message, int bar_size, int max_value = 1,
	      int update_step = 0, char bar_char = '=', std::ostream& out = std::cout);

  //! Sets the maximum value of the bar to \a max_value.
  void set_max(int max_value);

  //! Starts the bar, displaying the message.
  void start();

  //! Sets current value to maximum and finishes the progress.
  void finish();

  //! Increments bar value by one; this is thread-safe.
  ProgressBar& operator++();

  //! Increments bar value by \a val; this is thread-safe.
  ProgressBar& operator+=(int val);
};


} // namespace


#endif

// Local variables:
// mode: c++
// End:
