/*!
  \file qrdec.h

  \brief Provides the QR decomposition of a matrix.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __QRDEC_H__
#define __QRDEC_H__

#include <vector>
#include <iostream>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!  
  \brief Provides the QR decomposition of a matrix with entries of
  type \c T.

  The decomposition is computed using Householder reflectors with
  column pivoting. The algorithm is column-oriented, and the input
  matrix should be provided as a vector in column-major order.

  Scaling is performed to avoid overflow, the scaling factor should be
  taken into account when using the decomposition.
*/
template <class T>
class QRDecomposition {
  int _m, _n;             // Number of rows and columns
  T sfactor;              // Scaling factor
  std::vector<T> QR;      // Matrix containing Q and R
  std::vector<T> sigma;   // Reflector parameters & diagonal of R
  std::vector<T> msigma;  // The opposites of the sigma numbers
  std::vector<int> p;     // Pivot vector

public:
  /*!
    \brief Initializes an instance by computing the QR decomposition.

    \param A is the input matrix as a vector in column-major order.
    \param m number of rows in \a A.
    \param n number of columns in \a A.
    \param sing_eps is used during the process to decide the rank of
    \a A; during the decomposition, a column whose squared norm is
    less than \a sing_eps is considered to be zero.

    \throws ValueError if there are less rows than columns.
  */
  QRDecomposition(const std::vector<T>& A, int m, int n, double sing_eps = 1e-10)
    : _m(m), _n(n), QR(A), p(n)
  {
    if (m < n)
      throw ValueError(ERROR_MSG("Number of rows smaller than of columns"));
    
    /* Finds the scaling factor and scales the matrix */
    sfactor = 0;
    for (int i = 0; i < QR.size(); i++)
      if (NumberOperations::abs(QR[i]) > sfactor)
	sfactor = NumberOperations::abs(QR[i]);

    for (int i = 0; i < QR.size(); i++) QR[i] /= sfactor;

    /* Computes column squared norms */
    std::vector<T> cnorms(n, 0);

    for (int j = 0; j < n; j++)
      for (int i = 0; i < m; i++)
	cnorms[j] += QR[i + j * m] * QR[i + j * m];

    /* Does the QR decomposition with column pivoting */
    sigma.reserve(n);
    for (int j = 0; j < n; j++) p[j] = j;

    for (int k = 0; k < n; k++) {
      /* Pivoting step */
      int pivot = k;
      for (int j = k + 1; j < n; j++)
	if (cnorms[p[j]] > cnorms[p[pivot]])
	  pivot = j;

      if (cnorms[p[pivot]] <= sing_eps) break;  // We are finished

      { int tmp = p[k];
	p[k] = p[pivot];
	p[pivot] = tmp;
      }

      /* Computes the reflector */
      T s = NumberOperations::sqrt(cnorms[p[k]]);

      if (QR[k + p[k] * m] < 0) s *= -1;
      QR[k + p[k] * m] += s;

      T gamma_inv = s * QR[k + p[k] * m];
      sigma.push_back(s);

      /* Applies reflector to other columns */
      for (int j = k + 1; j < n; j++) {
	T tau = 0;
	for (int i = k; i < m; i++)
	  tau += QR[i + p[k] * m] * QR[i + p[j] * m];

	tau /= gamma_inv;
	for (int i = k; i < m; i++)
	  QR[i + p[j] * m] -= tau * QR[i + p[k] * m];
      }

      /* Updates column norms */
      for (int j = k + 1; j < n; j++)
	cnorms[p[j]] -= QR[k + p[j] * m] * QR[k + p[j] * m];
    }

    /* And we fill msigma */
    msigma = sigma;
    for (int i = 0; i < msigma.size(); i++)
      msigma[i] *= -1;
  }
  
  /*!
    \brief Returns the rank of the matrix, as determined by the QR
    decomposition.

    If the rank is \f$r\f$, then

    \f[
    R = \begin{pmatrix}
    R_{11}&R_{12}\\
    0&0
    \end{pmatrix}
    \f]

    where \f$R_{11}\f$ is an \f$r \times r\f$ matrix.
  */
  int rank() const { return sigma.size(); }

  /*!
    \brief Returns the original index of the \a j-th column after the
    pivoting.

    The decomposition is implemented with column pivoting; we have
    \f$AP = QR\f$, where \f$P\f$ is a permutation matrix that permutes
    the columns of \f$A\f$. So we have a decomposition for \f$AP\f$,
    which is a permuted version of \f$A\f$. The permutation is given
    as follows: the \a j-th column of \f$AP\f$ is the column of
    \f$A\f$ with index <tt>QR.perm(j)</tt>.
  */    
  int perm(int j) const { return p[j]; }

  /*!
    \brief Returns the scaling factor used.

    The decomposition is not computed for the given matrix \f$A\f$,
    but for \f$s^{-1} A\f$, where \f$s\f$ is the scaling factor used.
  */
  const T& scaling_factor() const { return sfactor; }

  /*! 
    \brief Returns entry <tt>(i, j)</tt> of matrix \f$R\f$ of the
    decomposition.
  */
  const T& R(int i, int j) const 
  { 
    if (i == j) return msigma[i];

    return QR[i + p[j] * _m]; 
  }

  //! Multiplies a vector \a x on the left by the matrix \f$Q\f$.
  void Q_mul(std::vector<T>& x) const
  {
    for (int k = sigma.size() - 1; k >= 0; k--) {
      T tau = 0;

      for (int i = k; i < _m; i++)
	tau += QR[i + p[k] * _m] * x[i];

      tau /= sigma[k] * QR[k + p[k] * _m];

      for (int i = k; i < _m; i++)
	x[i] -= tau * QR[i + p[k] * _m];
    }
  }

  //! Multiplies a vector \a x on the left by the matrix \f$Q^{\sf T}\f$.
  void Q_transpose_mul(std::vector<T>& x) const
  {
    for (int k = 0; k < sigma.size(); k++) {
      T tau = 0;

      for (int i = k; i < _m; i++)
	tau += QR[i + p[k] * _m] * x[i];

      tau /= sigma[k] * QR[k + p[k] * _m];

      for (int i = k; i < _m; i++)
	x[i] -= tau * QR[i + p[k] * _m];
    }
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
