/*!
  \file sdpcon.h

  \brief Container for constraints.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPCON_H__
#define __SDPCON_H__

#include <list>
#include <memory>
#include <sdpsl/sref.h>
#include <sdpsl/sdpenv.h>
#include <sdpsl/sparsecoeff.h>
#include <sdpsl/sdplincons.h>
#include <sdpsl/sdppolcons.h>
#include <sdpsl/sdpprogress.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A container for constraints of an SDP.

  Linear and polynomial constraints of a semidefinite programming
  problem, that is, instances of \c SDPLinearConstraint and \c
  SDPPolynomialConstraint, have to be processed before being used to
  generate a semidefinite programming problem.

  This processing is done by \c SDPConstraintContainer. An instance of
  this class accepts linear and polynomial constraints for
  processing. It transforms them into one or more instances of \c
  RawConstraint that it then stores. Later, the constraints contained
  in \c SDPConstraintContainer can be used to generate semidefinite
  programming problems.

  We also keep a list of all variables occurring in constraints added
  to the container. The values of environment parameters (cf. \c
  SDPEnvironment) regulate the behavior of \c
  SDPConstraintContainer. Namely, only numbers whose absolute value is
  larger than \c output_eps are considered nonzero.

  \note When a constraint is added to the container, the matrices
  provided as part of \c SDPLinearConstraint or \c
  SDPPolynomialConstraint are no longer necessary. They can be changed
  or destroyed without affecting the constraint added.

  \c SDPConstraintContainer supports shared references in the same way
  that \c Matrix does. See also \c shared_ref and \c const_shared_ref.

  \section sdpcon_par Paralellism

  Usually, constraint generation is the most computation-heavy part of
  generating an SDP. Multiple constraint containers can be created and
  used simultaneously (in different threads) to generate constraints
  without affecting each other. Then, the constraints generated in
  different containers can be added to the problem.

  \sa \c SDPProblem
*/  
class SDPConstraintContainer {
  struct Data {
    std::list<SDPVariable> variables;
    std::vector<bool> variables_contained;
    std::list<RawConstraint> constraints;
  };

  typedef std::shared_ptr<Data> data_ptr;

  int _num_ineqs;
  SDPEnvironment env;
  data_ptr data;

  /* Utility */
  void add_variable(const SDPVariable& X);

  void do_add_constraint(const SDPPolynomialConstraint& C, const SparseBasisCoefficients& sc, 
                         SDPProgressIndicator *progress);

  /* Mechanism for taking references */
  friend class shared_ref<SDPConstraintContainer>;
  friend class const_shared_ref<SDPConstraintContainer>;

  SDPConstraintContainer(const SDPConstraintContainer& other, const data_ptr& _data)
    : _num_ineqs(other._num_ineqs), env(other.env), data(_data) { }

  const data_ptr& take_data() const { return data; }

public:
  //! Creates an empty container with environment \a e.
  explicit SDPConstraintContainer(const SDPEnvironment& e = SDPEnvironment())
    : _num_ineqs(0), env(e), data(std::make_shared<Data>()) { }

  //! Copy constructor.
  SDPConstraintContainer(const SDPConstraintContainer& other)
    : _num_ineqs(other._num_ineqs), env(other.env), 
      data(std::make_shared<Data>(*other.data)) { }

  //! Deleted assignment operator.
  SDPConstraintContainer& operator=(const SDPConstraintContainer&) = delete;

  //! Returns \c true if the container is empty.
  bool empty() const { return data->constraints.empty(); }

  //! Erases all constraints in the container.
  void clear()
  {
    _num_ineqs = 0;
    data->variables.clear();
    data->variables_contained.clear();
    data->constraints.clear();
  }

  /*!
    \brief Returns the number of linear constraints in the container that
    are inequalities.
  */
  int num_ineqs() const { return _num_ineqs; }

  //! Returns a list of the variables used.
  const std::list<SDPVariable>& variables() const { return data->variables; }

  //! Returns constraints contained as raw constraints.
  const std::list<RawConstraint>& constraints() const { return data->constraints; }
  
  /* Operations */

  /*!
    \brief Adds a linear constraint to the container.
    
    \throws InfeasibleConstraint when the constraint added is detected
    to be infeasible. This happens for instance for an equality
    constraint when the left-hand side is always zero but the right-hand
    side is nonzero.
  */
  void add_constraint(const SDPLinearConstraint& C);

  /*!
    \brief Adds a polynomial constraint to the container using an
    automatically generated monomial basis.
    
    This function uses a monomial basis containing all monomials that
    occur in the constraint. Notice that this incurs in extra work,
    since one has to iterate through all polynomials in the constraint
    to assemble the monomial basis. It might save some time to
    assemble the monomial basis oneself.

    \sa \c SparseMonomialBasis
  */
  void add_constraint(const SDPPolynomialConstraint& C);

  /*!
    \brief Adds a polynomial constraint to the container using an
    automatically generated monomial basis and a progress indicator.

    This function uses a monomial basis containing all monomials that
    occur in the constraint. Notice that this incurs in extra work,
    since one has to iterate through all polynomials in the constraint
    to assemble the monomial basis. It might save some time to
    assemble the monomial basis oneself.
  */
  void add_constraint(const SDPPolynomialConstraint& C, SDPProgressIndicator& progress);

  /*!
    \brief Adds a polynomial constraint to the container using a sparse basis.
    
    About the use of bases for adding polynomial constraints, see \ref polybasics.
    
    \param C the polynomial constraint to be added.
    
    \param sc An object implementing the interface provided in \c
    SparseBasisCoefficients, which must be able to provide coefficients
    for each polynomial in the constraint.
    
    \throws InfeasibleConstraint when it is detected that the constraint
    being added is infeasible. This happens when the left-hand side is
    an identically zero polynomial, but the right-hand side is not.
    
    \throws InvalidBasis if some polynomial in the constraint cannot be
    processed by \a sc.
  */
  void add_constraint(const SDPPolynomialConstraint& C, const SparseBasisCoefficients& sc);

  /*!
    \brief Adds a polynomial constraint to the container using a sparse
    basis and a progress indicator.

    About the use of bases for adding polynomial constraints, see \ref polybasics.

    Adding polynomial constraints can be time-consuming. This function
    allows for a <em>progress indicator</em> object to be
    specified. This is an instance of a class derived from \c
    SDPProgressIndicator, that will be called to report the progress of
    the function.

    \param C the polynomial constraint to be added.

    \param sc An object implementing the interface provided in \c
    SparseBasisCoefficients, which must be able to provide coefficients
    for each polynomial in the constraint.

    \param progress A progress indicator.

    \throws InfeasibleConstraint when it is detected that the constraint
    being added is infeasible. This happens when the left-hand side is
    an identically zero polynomial, but the right-hand side is not.
  
    \throws InvalidBasis if some polynomial in the constraint cannot be
    processed by \a sc.
  */
  void add_constraint(const SDPPolynomialConstraint& C, const SparseBasisCoefficients& sc, 
                      SDPProgressIndicator& progress);
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
