/*!
  \file sdpenv.h

  \brief Environment parameters for \sdpsl.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPENV_H__
#define __SDPENV_H__

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Keeps environment parameters for \sdpsl.

  Some parameters regulate how different classes of \sdpsl
  behave. These are:

  - \c output_eps: during constraint generation, only numbers whose
    absolute value are greater than \c output_eps are considered
    nonzero. Default value: 1e-15.

  - \c tolerance_eps: used by functions when numerical errors are
    concerned (for instance, during matrix inversion). Default value: 1e-15.

  These parameters are kept in the \e environment of \sdpsl. Instances
  of \c SDPEnvironment can be used to access such values.

  There are two kinds of instances of \c SDPEnvironment: those that
  access the default values of the parameters, and those that contain
  custom values. The constructor of \c SDPEnvironment that takes no
  parameters creates an instance accessing the default values. These
  values can be changed globally through the appropriate static
  functions of \c SDPEnvironment; once these values are changed, every
  instance of \c SDPEnvironment accessing the default values will also
  change.

  The constructor that accepts arguments allows for custom values to
  be defined. Such instances of \c SDPEnvironment ignore the default
  values. Here is an example of how it works:

  \include env.cpp

  Classes such as \c SDPProblem and \c SDPConstraintContainer take
  instances of \c SDPEnvironment during construction to have access to
  the parameters.

  \todo 
  The current implementation is very simple and either all
  variables are custom, or all variables are default. In the future it
  might be interesting to solve this issue.
*/
class SDPEnvironment {
  static double global_output_eps;
  static double global_tolerance_eps;

  double _output_eps, _tolerance_eps;
  
public:
  //! Creates an environment containing the default values of the parameters.
  SDPEnvironment()
    : _output_eps(-1), _tolerance_eps(-1) { }

  /*!
    \brief Creates a custom environment.

    \throws ValueError if either parameter passed is negative.
  */
  SDPEnvironment(double out_eps, double tol_eps);

  /*!
    \brief Sets the default value of \c output_eps.

    \throws ValueError if the value is negative.
  */
  static void set_output_eps(double eps);

  /*!
    \brief Sets the default value of \c tolerance_eps.

    \throws ValueError if the value is negative.
  */
  static void set_tolerance_eps(double eps);

  //! Returns the value of \c output_eps.
  double output_eps() const { return _output_eps < -0.5 ? global_output_eps : _output_eps; }

  //! Returns the value of \c tolerance_eps.
  double tolerance_eps() const { return _tolerance_eps < -0.5 ? global_tolerance_eps : _tolerance_eps; }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
