/*!
  \file sdpproblem.h

  \brief A class representing a semidefinite programming problem.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPPROBLEM_H__
#define __SDPPROBLEM_H__

#include <list>
#include <sdpsl/sdpenv.h>
#include <sdpsl/sdpcon.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Represents a semidefinite programming problem.

  A semidefinite programming problem consists of an objective function
  and a collection of constraints. This class keeps the objective
  function of the problem together with references to one or more
  instances of \c SDPConstraintContainer that contain the constraints
  of the problem.

  The value of the environment parameter \c output_eps is used to
  determine which numbers of the objective function are nonzero (see
  \c SDPEnvironment). Namely, only numbers whose absolute value are
  greater than \c output_eps are considered nonzero.

  Since we keep shared references to instances of \c
  SDPConstraintContainer, adding a constraint container to the problem
  is an efficient, constant-time operation, and it is guaranteed that
  the constraints will survive even if the original constraint
  container is destroyed. Notice however that further changes to
  instances of \c SDPConstraintContainer that have been added to the
  problem will also alter the problem, since the data is not copied.

  \sa \c shared_ref, \c const_shared_ref, \c SDPConstraintContainer
*/
class SDPProblem {
  SDPEnvironment env;
  std::list<SDPVariable> _obj_variables;
  MatrixEntryList _objective_entries;
  std::list<const_shared_ref<SDPConstraintContainer> > cc_list;

public:
  //! Constructs an empty problem with given environment.
  explicit SDPProblem(const SDPEnvironment& e = SDPEnvironment()) : env(e) { }

  /*!
    \brief Sets the objective function of the problem.

    The instance \a C of \c SDPLinearCombination associates numerical
    matrices do variables. This association is used as the objective
    function.

    If an objective function was already set, then it is
    overwritten. After the objective function is set, further changes to
    \a C do not alter the objective function.
  */
  void set_objective(const SDPLinearCombination& C);

  /*!
    \brief Adds a constraint container \a cc to the problem. This uses
    shared references (see detailed description of \c SDPProblem).
  */
  SDPProblem& operator<<(const SDPConstraintContainer& cc);

  //! Returns the list of variables occurring in the objective function.
  const std::list<SDPVariable>& objective_variables() const { return _obj_variables; }

  /*!  
    \brief Returns a list of all variables in the problem. This
    includes those of the objective function, and those of the
    constraints.
  */
  std::list<SDPVariable> variables() const;

  //! Returns the list of entries of the objective function.
  const MatrixEntryList& objective_entries() const { return _objective_entries; }

  //! Returns a list of references to the containers that were added.
  const std::list<const_shared_ref<SDPConstraintContainer> >& containers() const { return cc_list; }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
