/*!
  \file sdpprogress.h

  \brief Base class for progress indicators.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPPROGRESS_H__
#define __SDPPROGRESS_H__

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Base class for progress indicators.

  Some operations during the generation of the problem, like the
  adding polynomial constraints, can be time-consuming. In such cases
  a class derived from \c SDPProgressIndicator can be supplied so that
  the caller can be kept up-to-date with the progress of the
  operation. One use of this is to give the user a graphical
  indication of progress (cf. \c ProgressBar).
*/
class SDPProgressIndicator {
public:
  //! Virtual destructor.
  virtual ~SDPProgressIndicator() { }

  /*! 
    \brief Sets the maximum value that the progress indicator will attain.

    This method is called before any other method; \a max_value
    contains the maximum value that the progress indicator will
    attain.
  */
  virtual void set_max(int max_value) = 0;

  //! Called after \c set_max; indicates that the operation is starting.
  virtual void start() = 0;

  //! Called after the operation is concluded.
  virtual void finish() = 0;

  //! Increases the progress indicator by one.
  virtual SDPProgressIndicator& operator++() = 0;

  //! Increases the progress indicator by \a val.
  virtual SDPProgressIndicator& operator+=(int val) = 0;
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
