/*!
  \file sdpvar.h

  \brief A variable in a semidefinite programming problem.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPVAR_H__
#define __SDPVAR_H__

#include <vector>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A variable in a semidefinite programming problem.

  Any instance of this class represents a variable in a semidefinite
  programming problem. Variables have indices, attributed uniquely to
  them at creation, and two variables are the same if their indices
  coincide. Each variable also holds its associated size.
*/
class SDPVariable {
  static std::vector<int> variable_sizes;

  int _idx;
  int _size;

public:
  //! Creates a new variable, given its size.
  explicit SDPVariable(int size) 
    : _idx(variable_sizes.size()), _size(size) { variable_sizes.push_back(size); }

  //! Returns total number of variables.
  static int num_variables() { return variable_sizes.size(); }
  
  //! Returns size of a variable given its index.
  static int variable_size(int idx) { return variable_sizes[idx]; }

  int idx() const { return _idx; }  //!< Returns the variable index.
  int size() const { return _size; }  //!< Returns the variable size.

  /*!
    \brief Compares two variables according to their indices.

    Returns true if the index of this variable is less than the index
    of the \a other variable.
  */
  bool operator<(const SDPVariable& other) const { return _idx < other._idx; }

  //! Returns true if the variables are the same (have the same index).
  bool operator==(const SDPVariable& other) const { return _idx == other._idx; }

  //! Returns true if the variables are different (have different indices).
  bool operator!=(const SDPVariable& other) const { return _idx != other._idx; }
};

}  // namespace


#endif

// Local variables:
// mode: c++
// End:
