/*!
  \file sparsecoeff.h

  \brief Base class for sparse polynomial bases.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SPARSECOEFF_H__
#define __SPARSECOEFF_H__

#include <vector>
#include <sdpsl/polynomial.h>
#include <sdpsl/vectormask.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!  
  \brief Virtual base class of objects used to specify a basis for
  polynomial constraints.

  Different bases of polynomials can be used when adding a polynomial
  constraint to a constraint container. This is done by classes
  derived from \c SparseBasisCoefficients who implement the virtual
  functions below.

  \sa \ref otherbases. For examples of classes derived from \c
  SparseBasisCoefficients, see \c SparseMonomialBasis and \c
  SparseBasisChanger.

  \par "Why sparse?"  This class is called "sparse" because it uses
  masks to tell which of the basis elements are used in the expansion
  of a given polynomial. One might work with a large space of
  polynomials, but each polynomial might have a concise representation
  on the chosen basis (e.g., if one considers the standard monomial
  basis, then polynomials with few monomials have small
  representations on the standard monomial basis).

  \todo
  In the future, add a dense version of this class, that ignores the masks.
*/
class SparseBasisCoefficients {
  
public:
  //! Virtual destructor
  virtual ~SparseBasisCoefficients() { }

  //! Returns the size of the basis.
  virtual int basis_size() const = 0;

  /*! 
    \brief Returns the expansion of a given polynomial on the basis.

    Given a polynomial \a p, a vector \a coeffs, and a \c VectorMask
    \a mask, this function should compute the expansion of the
    polynomial \a p on the basis it represents and put this expansion
    on the vector \a coeffs, whose size should be at least equal to
    the size of the basis. Here, \a mask is a bit-mask that specifies
    which positions of the vector \a coeffs are nonzero.

    It is assumed that the vector and the mask have the size of the
    basis. The mask need not be clear.

    \returns \c true if the expansion was successfully computed; \c
    false otherwise.
  */
  virtual bool basis_coefficients(const Polynomial& p, std::vector<Number>& coeffs,
                                  VectorMask& mask) const = 0;
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
