"""
A function to create an SDP that applies Fan's theorem to a given matrix.

AUTHOR:

- Fernando M. de Oliveira Filho <fmario AT gmail.com>
"""

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

from sage.sdpsl import *

def fans_theorem(A, k):
    """
    Return an SDPSL problem whose optimal value is the sum of the ``k``
    largest eigenvalues of a symmetric matrix ``A``.
    """
    if not A.is_square():
        raise ValueError('A must be a square matrix')
        
    n = A.nrows()

    # Variables and the constraint container
    X, Y = SDPVariable(n), SDPVariable(n)
    cc = SDPConstraintContainer()

    # First the trace constraint
    cc.add_constraint(identity_matrix(RR, n) * X == k)

    # Then the correspondence between X and Y
    E = SDPMatrix(n, 1)
    mask = SDPMatrixMask(n)
    C = (E, mask) * X + (E, mask) * Y == 0

    for i in xrange(n):
        for j in xrange(i, n):
            if i == j: 
                C.set_rhs(1)
            else:
                C.set_rhs(0)

            mask[i, j] = True
            cc.add_constraint(C)
            mask[i, j] = False

    # Then we create the problem and return it
    problem = SDPProblem()

    problem.set_objective(A * X)
    problem.add_container(cc)

    return problem


# Local variables:
# mode: python
# End:
