"""
A semidefinite programming approach to compute the Cohn-Elkies
sphere-packing bounds.

AUTHOR:

- Fernando M. de Oliveira Filho <fmario AT gmail.com>
"""

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

from sage.sdpsl import *

def laguerre_polys(a, d, PR, x):
    """
    Return a list of the first few Laguerre polynomials.

    This function returns a list with the polynomials L_k^a(x) for k =
    0, ..., d. The polynomials are elements of the polynomial ring PR.
    """
    
    ret = [ PR(1), -x + a + 1 ]
    for k in xrange(2, d + 1):
        ret.append((2 + (a - 1 - x) / k) * ret[-1]
                   - (1 + (a - 1) / k) * ret[-2])

    return ret

class SpherePacking:
    """
    An implementation of a semidefinite programming approach to
    compute approximations for the Cohn-Elkies linear programming
    bound for sphere packings.
    """
    
    def __init__(self, n, d, moment_monomials = False, encode_monomials = False, 
                 prec = 256, csdp = 'csdp'):
        """
        Initialize data for generating the Cohn-Elkies bounds.

        INPUT:

        - ``n`` -- is the dimension of the space; an integer >= 2.

        - ``d`` -- is the maximum degree of polynomials used; an odd integer.

        - ``moment_monomials`` -- if True, then the standard monomial
          basis is used for the moment matrix; if False, a basis of
          normalized Laguerre polynomials is used.

        - ``encode_monomials`` -- if True, then the standard monomial
          basis is used to encode the polynomial identities; if False,
          then a basis of Laguerre polynomials is used.
        
        - ``prec`` -- is the precision to be used, in bits.

        - ``csdp`` -- command to run CSDP solver.
        """
        
        if int(n) <= 1:
            raise ValueError('n must be >= 2')
        
        if int(d) % 2 == 0:
            raise ValueError('d must be odd')

        self.n = int(n)
        self.d = int(d)
        self.pi = RealField(prec).pi()  # An approximation for pi
        self.csdp = csdp

        sdpsl_set_gmp_prec(prec)
        
        # The polynomial ring
        self.PR = PolynomialRing(RealField(prec), 'x')
        x = self.PR.gen()
        
        # Bases of monomials and polynomials that we work with
        self.mon_basis = [ x^(2 * k) for k in xrange(d + 1) ]
        
        tmp = laguerre_polys(n / 2.0 - 1, d, self.PR, 2 * self.pi * x^2)
        self.lag_basis = [ p / max(abs(c) for c in p.coefficients()) for p in tmp ]

        # Computes polynomials for Fourier inversion
        tmp = laguerre_polys(n / 2.0 - 1, d, self.PR, self.pi * x^2)
        self.fourier_polys = [ factorial(k) * self.pi^(-k) * tmp[k] for k in xrange(len(tmp)) ]

        # The moment matrix
        if moment_monomials:
            v = vector(self.PR, self.mon_basis[:floor(d / 2) + 1])
        else:
            v = vector(self.PR, self.lag_basis[:floor(d / 2) + 1])
            
        self.V = v.column() * v.row()

        # The matrices of Fourier inverses
        def inverse(p):
            ret = 0
            for c, k in zip(p.coefficients(), p.exponents()):
                ret += c * self.fourier_polys[k / 2]

            return ret
        
        self.F = matrix(self.PR, self.V.nrows(), self.V.ncols(), lambda i, j: inverse(self.V[i, j]))
        self.x2_F = matrix(self.PR, self.V.nrows(), self.V.ncols(), lambda i, j: inverse(x^2 * self.V[i, j]))

        # Variables of the problem
        self.S0 = S0 = SDPVariable(floor(d / 2) + 1)
        self.S1 = S1 = SDPVariable(floor(d / 2) + 1)
        self.Q = Q = SDPVariable(floor(d / 2) + 1)        

        # Container for the polynomial constraint and the constraint itself
        self.poly_cc = SDPConstraintContainer()
        
        self.poly_cc.add_constraint(self.V(0) * S0 == 2^(-n))

        if encode_monomials:
            self.poly_cc.add_constraint(self.F * S0 + self.x2_F * S1 + (x^2 - 1) * self.V * Q == 0)
        else:
            try:
                changer = SDPSparseBasisChanger(self.mon_basis, self.lag_basis)
            except ValueError as ve:
                raise ValueError(ve.message + '\nTry decreasing tolerance_eps with the command: sdpsl_set_tolerance_eps')
                
            self.poly_cc.add_constraint(self.F * S0 + self.x2_F * S1 + (x^2 - 1) * self.V * Q == 0, 
                                        changer)

    def make_objective_function(self, fac = 1):
        """
        Return a linear combination containing the objective function
        of the problem, multiplied by ``fac``.
        """
        
        return fac * self.F(0) * self.S0 + fac * self.x2_F(0) * self.S1

    def make_opt_problem(self):
        """
        Return a problem whose optimal value is an approximation for
        the Cohn-Elkies bound.
        """
        
        problem = SDPProblem()
        problem.set_objective(self.make_objective_function(-1))
        problem.add_container(self.poly_cc)

        return problem

    def make_feas_problem(self, eps, solution = None, verbose = False):
        """
        Return a pure feasibility problem attempting to find a good
        solution for the Cohn-Elkies bounds.

        Given a number ``eps``, and a solution to the optimization
        problem that approximates the Cohn-Elkies bound, we generate a
        pure feasibility problem in which the objective function is
        absent, but added as a constraint, so that the objective
        function computed a the solution of the feasibility problem is
        at most the objective value plus eps. So we allow for
        suboptimal solutions, hoping to get strictly feasible
        solutions.

        INPUT:

        - ``eps`` -- the tolerance parameter.

        - ``solution`` -- a solution to the Cohn-Elkies bound
          optimization problem; if ``None``, then a solution is
          computed using CSDP.

        - ``verbose`` -- if True, then the CSDP output (if generated)
          is shown.
        """
        
        if solution is None:
            problem = self.make_opt_problem()
            S = SDPSolution(problem.variables())
            
            solver = CSDPSolver(self.csdp)
            ret = solver.solve(problem, solution = S, verbose = verbose)

            if ret.ret_code != 0 and ret.ret_code != 3:
                raise RuntimeError('CSDP failed to solve the problem: return code = %d' % ret.ret_code)
        elif isinstance(solution, SDPSolution):
            S = solution
        else:
            raise RuntimeError('A solution is expected')

        C = self.make_objective_function(1)
        obj_value = C.evaluate_at_solution(S)

        cc = SDPConstraintContainer()
        cc.add_constraint(C <= obj_value + eps)

        problem = SDPProblem()
        problem.add_container(cc)
        problem.add_container(self.poly_cc)

        return problem

    def project(self, solution):
        """
        Projects a solution onto the affine subspace generated by the
        constraints of the problem, returning the projected solution.
        """
        return project_solution([ self.poly_cc ], solution)

    def _show_info(self, S, obj):
        print 'Minimum eigenvalues...'
        print '  of S0 =', min(sdpsl_double_eigs(S[self.S0]))
        print '  of S1 =', min(sdpsl_double_eigs(S[self.S1]))
        print '  of Q  =', min(sdpsl_double_eigs(S[self.Q]))
        print ''

        print 'Maximum constraint violation =', max(feasibility_measures([ self.poly_cc ], S))
        print 'Objective value              =', obj.evaluate_at_solution(S)

    def run_example(self, eps):
        """
        Runs an example computation of the Cohn-Elkies bound.

        A call to this method uses CSDP to compute the optimal value
        of the optimization SDP for the Cohn-Elkies bound, then
        generates a feasibility version of the problem using the given
        number eps, then projects the solution. Information about the
        three solutions found is displayed.
        """
        solver = CSDPSolver(self.csdp)

        # Solve the optimization problem
        problem = self.make_opt_problem()
        S = SDPSolution(problem.variables())

        ret = solver.solve(problem, solution = S)

        if ret.ret_code != 0 and ret.ret_code != 3:
            raise RuntimeError('CSDP failed to solve the problem: return code = %d: %s' % (ret.ret_code, ret.msg))
        
        # Solve the feasibility problem
        feas = self.make_feas_problem(eps, solution = S)
        T = SDPSolution(feas.variables())
        
        ret = solver.solve(feas, solution = T)

        if ret.ret_code != 0 and ret.ret_code != 3:
            raise RuntimeError('CSDP failed to solve the problem: return code = %d: %s' % (ret.ret_code, ret.msg))

        # Projects the solution
        U = self.project(T)

        # Shows info on S, T, and U
        obj = self.make_objective_function(1)
        
        print 'Info on optimal solution'
        print '------------------------'
        print ''

        self._show_info(S, obj)

        print ''
        print 'Info on feasibility solution'
        print '----------------------------'
        print ''

        self._show_info(T, obj)
        
        print ''
        print 'Info on projected solution'
        print '--------------------------'
        print ''

        self._show_info(U, obj)

# Local variables:
# mode: python
# End:
