"""
A program to compute the semidefinite programming bound of D. de Laat,
F.M. de Oliveira Filho, and F. Vallentin for the densities of multiple size
spherical cap packings.

REFERENCES:

- D. de Laat, F.M. de Oliveira Filho, and F. Vallentin, Upper bounds
  for packings of spheres of several radii, arXiv:1206.2608, 2012, 31pp.

AUTHOR:

- David de Laat

NOTE:

This program requires SDPA-QD.
"""

from sage.sdpsl import *
from subprocess import call


### Code for the geometric bound by Florian ###

def triangle_angle(r1, r2, r3):
    return acos((cos(r1) - cos(r2) * cos(r3)) / (sin(r2) * sin(r3)))
    
def triangle_area(r1, r2, r3):
    t1 = triangle_angle(r1, r2, r3)
    t2 = triangle_angle(r2, r1, r3)
    t3 = triangle_angle(r3, r1, r2)
    return t1 + t2 + t3 - pi
    
def height(angle):
    return 1 - cos(angle)

def cap_area(angle):
    return 2 * pi * height(angle)

def D(r1, r2, r3):
    at = triangle_area(r1+r2, r1+r3, r2+r3)
    a1 = triangle_angle(r2+r3, r1+r2, r1+r3) * cap_area(r1)
    a2 = triangle_angle(r1+r3, r2+r3, r1+r2) * cap_area(r2)
    a3 = triangle_angle(r1+r2, r1+r3, r2+r3) * cap_area(r3)
    return (a1 + a2 + a3) / (2 * pi * at)
    
def florianbound(angles):
    m = None
    for i in range(len(angles)):
        for j in range(i, len(angles)):
            for k in range(j, len(angles)):
                d = D(angles[i], angles[j], angles[k])
                if m == None or d > m:
                    m = d
    return m
    

### Code for our semidefinite programming bound ###

def sphere_area(n=3):
    return 2 * pi ** (n / 2) / gamma(n / 2)    

def normalized_spherical_cap_area(angle, n = 3):
    alpha = (n - 3) / 2
    z = var('u')
    v = integrate((1 - z^2)^alpha, z, cos(angle), 1)
    return sphere_area(n - 1) / sphere_area(n) * v

def sdpbound(angles, n, d, prec=128):
    sdpsl_set_gmp_prec(prec)
    sdpsl_set_output_eps(1e-40)
    sdpsl_set_tolerance_eps(1e-30)
    RF = RealField(prec)
    PR = PolynomialRing(RF, 'x')
    x = PR.gen()
    
    N = len(angles)
    w = [RF(normalized_spherical_cap_area(angle, n)) for angle in angles]
    
    f = [SDPVariable(N) for i in xrange(2*d+1)]    
    Q = [[SDPVariable(d+1) for i in xrange(N)] for j in xrange(N)]
    R = [[SDPVariable(d) for i in xrange(N)] for j in xrange(N)]
    s = SDPVariable(1)

    cc = SDPConstraintContainer()

    J = SDPMatrix(N, initial_value = 1)
    mask = SDPMatrixMask(N)
    for i in range(N):
        mask[i, i] = True
        comb = -1 * s
        for k in range(2*d+1):
            comb += (J, mask) * f[k]
        cc.add_constraint(comb <= -w[i])
        mask[i, i] = False
    
    v = vector(PR, [x^k for k in xrange(d+1)])
    V = v.column() * v.row()
    u = vector(PR, [x^k for k in xrange(d)])
    
    for i in xrange(N):
        for j in xrange(N):
            if not(i == 0 and j == 1):
                U = (x + 1) * (RF(cos(angles[i] + angles[j])) - x) * u.column() * u.row()
                comb = V * Q[i][j] + U * R[i][j]
                for k in xrange(2*d+1):
                    M = SDPPolyMatrix(N, PR)
                    M[i, j] = jacobi_P(k, (n-3)/2, (n-3)/2, x) / jacobi_P(k, (n-3)/2, (n-3)/2, 1)
                    comb += M * f[k]
                cc.add_constraint(comb == -sqrt(w[i]*w[j]))

    problem = SDPProblem()
    problem.set_objective(-1 * s)
    problem.add_container(cc)
    problem.write_sdpa('out.sdpa', prec)
    call(['sdpa_qd', '-ds', 'out.sdpa', '-o', 'res', '-pt', '2'])


### Examples from the paper ###

# Truncated octahedron
angles = [asin(1/3), asin(sqrt(1/3))]
print RR(6*normalized_spherical_cap_area(angles[0]) + 8*normalized_spherical_cap_area(angles[1]))
sdpbound(angles, 3, 12)
print RR(florianbound(angles))
print RR(19*normalized_spherical_cap_area(angles[0]) + 4*normalized_spherical_cap_area(angles[1]))

# 5-Prism
angles = [pi/5, 3*pi/10]
print RR(5*normalized_spherical_cap_area(angles[0]) + 2*normalized_spherical_cap_area(angles[1]))
sdpbound(angles, 3, 12)

# Local variables:
# mode: python
# End:
