#
# Wrapper for SDPSolutionAnalyzer from SDPSL
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

def feasibility_measures(containers, solution):
    """
    feasibility_measures(containers, solution)
    
    Return measures of violation for a solution and constraints.

    Given a linear constraint and a solution, its measure of violation
    depends on its type:

    - For == constraints, it equals the absolute value of the
      difference of the right-hand side and the left-hand side.

    - For <= constraints, it equals the maximum of zero and the
      left-hand side minus the right-hand side.

    - For >= constraints, it equals the maximum of zero and the
      right-hand side minus the left-hand side.

    This means that the measure of violation is always a nonnegative
    number, and the more positive it is the more violated the
    constraint is.

    Measures of violation can also be applied to polynomial
    constraints, since they are transformed into linear constraints
    when added to a container.

    INPUT:

    - ``containers`` -- a list of containers with the constraints that
      will be tested.

    - ``solution`` -- the solution that will be tested.

    The function returns a list with the violation measures of each
    constraint in the containers given, in the order they are given.
    """
    if not containers:
        return [ ]
    
    cdef SDPConstraintContainer_wrapper cc_wrapper
    cdef SDPMatrixSolution_wrapper sol_wrapper
    cdef int N
    cdef MPFRNumber *ret_measures

    # Assembles list of containers
    n = len(containers)
    cdef SDPSLConstraintContainer *cc_list

    cc_list = <SDPSLConstraintContainer *> malloc((n + 1) * cython.sizeof(SDPSLConstraintContainer))

    if cc_list is NULL:
        raise MemoryError('Failed to allocate memory for cc_list')

    for k from 0 <= k < n:
        cc = containers[k]
        
        if not isinstance(cc, SDPConstraintContainer):
            free(cc_list)
            raise ValueError('List of containers must contain instances of SDPConstraintContainer')

        cc_wrapper = cc._container
        cc_list[k] = cc_wrapper._container

    cc_list[n] = NULL

    # Calls SDPSL
    tmpsol = solution._get_sdp_matrix_map()
    sol_wrapper = tmpsol._solmap

    try:
        sig_on()
        sdpsl_feasibility_measures(cc_list, sol_wrapper._solmap,
                                   cython.address(N), cython.address(ret_measures))
    finally:
        free(cc_list)
        sig_off()    

    # Assembles the measures for SAGE
    RF = RealField(SDPEnvironment.gmp_prec)
    tmp = RealNumber(RF, 0)
    measures = [ ]
    
    for k from 0 <= k < N:
        mpfr_set(tmp.value, (<mpfr_t *> ret_measures)[k], <mpfr_rnd_t> 0)
        measures.append(RealNumber(RF, tmp))

    free(ret_measures)
        
    return measures


def sdpsl_double_eigs(A, mask = None):
    """
    sdpsl_double_eigs(A, mask = None)

    Return the eigenvalues of symmetric matrix ``A``.

    This function computes the eigenvalues of an SDPSL matrix using
    LAPACK.
    
    INPUT:

    - ``A`` -- a symmetric SDPSL matrix of numbers, i.e., an instance
      of :class:`SDPMatrix`. It is assumed that ``A`` is symmetric but
      it is not checked; in any case only the lower-diagonal part is used.

    - ``mask`` -- is an instance of :class:`SDPMatrixMask` that tells
      which positions are to be considered nonzero. If equal to
      ``None``, then all positions are potentially nonzero.

    The function returns a list with the computed eigenvalues.

    EXCEPTIONS:

    - ``RuntimeError`` -- if call to LAPACK fails.
    """
    if not isinstance(A, SDPMatrix):
        raise ValueError('Matrix must be instance of SDPMatrix')

    if not mask is None and not isinstance(mask, SDPMatrixMask):
        raise ValueError('Mask must be instance of SDPMatrixMask')

    if not A.is_square():
        raise ValueError('Matrix must be square')

    if not mask is None:
        if A.nrows() != mask.nrows() or A.ncols() != mask.ncols():
            raise ValueError('Sizes of matrix and mask must coincide')

    cdef SDPMatrix_wrapper mat_wrapper = A._matrix
    cdef SDPMatrixMask_wrapper mask_wrapper
    cdef double *eigs

    try:
        sig_on()

        if mask is None:
            ret = sdpsl_double_eigenvalues(mat_wrapper._matrix, NULL, cython.address(eigs))
        else:
            mask_wrapper = mask._mask
            ret = sdpsl_double_eigenvalues(mat_wrapper._matrix, mask_wrapper._mask, 
                                           cython.address(eigs))
    finally:
        sig_off()

    if ret:
        raise RuntimeError('LAPACK call failed')

    # Assembles the result
    RF = RealField(53)
    ret_eigs = [ ]
    n = A.nrows()
    
    for k from 0 <= k < n:
        ret_eigs.append(RealNumber(RF, eigs[k]))

    free(eigs)

    return ret_eigs
    

def project_solution(containers, solution, result = None):
    """
    project_solution(containers, solution, result = None)
    
    Project a solution of an SDP onto the affine subspace given by
    some constraints.

    One way to improve the feasibility measures of a solution for some
    constraints is to project this solution onto the affine subspace
    generated by these constraints. This is done by the
    ``project_solution`` function.

    INPUT:

    - ``containers`` -- a list of constraint containers with the
      constraints that generate the affine subspace onto which the
      solution is to be projected. Only the equality constraints in
      the containers are considered. The system for the projection
      cannot be overdetermined, i.e., it must have at least as many
      variables as equalities.

    - ``solution`` -- the solution to be projected.

    OUTPUT:

    - ``result`` -- an instance of :class:`SDPSolution` that will
      contain the result of the projection. The solution is adapted as
      needed, i.e., new variables are added if necessary. If
      ``result`` is ``None``, then a new instance of
      :class:`SDPSolution` is created.

    The function returns the projected solution. 

    EXCEPTIONS:

    - ``ValueError`` -- If the system given by the constraints is
      overdetermined or infeasible. To decide whether the system is
      infeasible, environment parameter ``tolerance_eps`` is used (see
      :ref:`sdpenvironment`).

    .. note::

       If after a projection the quality of a solution (as determined,
       for instance, by a call to :func:`feasibility_measures`) is
       worse than before the projection, then most likely you are
       running into numerical trouble. It is always good to work with
       a higher precision for GMP in such cases.
    """
    if not containers:
        raise ValueError('No constraint containers specified')
    
    # We create a set of all variables in the containers and the solution
    variables = [ ]
    for cc in containers:
        if not isinstance(cc, SDPConstraintContainer):
            raise ValueError('Constraint container must be instance of SDPConstraintContainer')

        variables += cc.variables()

    if not isinstance(solution, SDPSolution):
        raise ValueError('Solution must be instance of SDPSolution')

    variables += solution.variables()

    var_set = set(variables)

    # We add the variables not in result to the result. If result is
    # None, then we actually create it.
    if result is None:
        result = SDPSolution()

    if not isinstance(result, SDPSolution):
        raise ValueError('Resulting solution must be instance of SDPSolution')
        
    for X in var_set:
        if not X in result:
            result.create_matrix([ X ])

    # We assemble the array of containers
    cdef SDPConstraintContainer_wrapper cc_wrapper
    cdef SDPSLConstraintContainer *cc_list

    n = len(containers)
    cc_list = <SDPSLConstraintContainer *> malloc((n + 1) * cython.sizeof(SDPSLConstraintContainer))

    if cc_list is NULL:
        raise MemoryError('Failed to allocate memory for cc_list')

    for k from 0 <= k < n:
        cc_wrapper = containers[k]._container
        cc_list[k] = cc_wrapper._container

    cc_list[n] = NULL

    # Calls SDPSL
    cdef SDPMatrixSolution_wrapper sol_wrapper, res_wrapper
    
    tmpsol = solution._get_sdp_matrix_map()
    sol_wrapper = tmpsol._solmap

    tmpres = result._get_sdp_matrix_map()
    res_wrapper = tmpres._solmap

    try:
        sig_on()
        ret = sdpsl_gmp_project(cc_list, sol_wrapper._solmap, res_wrapper._solmap)
    finally:
        free(cc_list)
        sig_off()    

    if ret == VALUE_ERROR:
        raise ValueError('System is overdetermined')
    elif ret == INFEASIBLE_SYSTEM:
        raise ValueError('System is infeasible with tolerance_eps = %s'
                         % str(SDPEnvironment.tolerance_eps))
        
    result._set_from_matrix_map(tmpres)

    return result
    

# Local variables:
# mode: python
# End:

