#
# SDPSL SDPConstraintContainer
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

cdef class SDPConstraintContainer_wrapper:
    """
    Wrapper for SDPConstraintContainer from SDPSL.
    """
    cdef SDPSLConstraintContainer _container
    
    def __cinit__(self):
        self._container = create_constraint_container()

    def __dealloc__(self):
        destroy_constraint_container(self._container)

    def add_linear_constraint(self, constraint):
        cdef SDPLinearConstraint_wrapper tmp

        try:
            sig_on()
            tmp = constraint._constraint
            ret = cc_add_linear(self._container, tmp._constraint)
        finally:
            sig_off()

        return ret

    def add_poly_constraint(self, constraint, changer = None, progress = False, progress_msg = None):
        cdef SDPPolynomialConstraint_wrapper tmp_con = constraint._constraint
        cdef SDPSparseBasisChanger_wrapper tmp_changer
        
        if progress_msg is None:
            progress_msg = 'Adding constraint'

        if changer is None:
            try:
                sig_on()
                ret = cc_add_poly(self._container, tmp_con._constraint, progress, progress_msg)
            finally:
                sig_off()
        else:
            try:
                sig_on()
                tmp_changer = changer._changer
                ret = cc_add_poly_with_basis(self._container, tmp_con._constraint,
                                             tmp_changer._changer, progress, progress_msg)
            finally:
                sig_off()

        return ret
    
    def num_constraints(self):
        return cc_get_num_constraints(self._container)

    def clear(self):
        cc_clear(self._container)

    
class SDPConstraintContainer:
    """
    An SDPSL constraint container.

    Linear and polynomial constraints, that is, instances of
    :class:`SDPLinearConstraint` and :class:`SDPPolynomialConstraint`,
    have to be converted to a "raw" format in order to be used in
    SDPSL problems. This operation is done with constraint containers.

    Constraints can be added to constraint containers, and these can
    be later added to problems and used in other ways.

    .. note::

       When a constraint is added, its data is copied, and further
       changes to the constraint matrices do not change the data
       inside the constraint container.

    Constraint containers are SDPSL's way of supporting parallelism in
    constraint generation. Multiple containers can be created and
    constraints can be added to different containers in different
    threads.
    
    EXAMPLES:

    Adding a linear constraint::

        sage: X = SDPVariable(5)
        sage: A = random_matrix(RR, 5)
        sage: cc = SDPConstraintContainer()
        sage: cc.add_constraint(A * X == 1)
        sage: cc
        SDPSL constraint container with 1 raw constraint(s)

    If the constraint can be seen to be infeasible, an exception is
    raised::

        sage: cc.add_constraint(matrix(RR, 5) * X == 2)
        ValueError: Constraint is infeasible

    Polynomial constraints can be added in a similar way, and then the
    standard monomial basis is used for the representation (see also
    :ref:`addpoly`)::

        sage: PR = PolynomialRing(RR, 'x')
        sage: x = PR.gen()
        sage: p = x^2 + 2
        sage: A = random_matrix(PR, 5)
        sage: cc.add_constraint(A * X == p)
        sage: cc
        SDPSL constraint container with 4 raw constraint(s)

    Notice that the number of raw constraints went from 1 to
    4. Polynomial constraints are transformed to linear constraints
    when added to a constraint container, the same constraint
    originates several linear constraints.

    Other bases (see :ref:`otherbases`) can be used to represent polynomial
    constraints. An arbitrary polynomial basis can be specified using
    an instance of :class:`SDPSparseBasisChanger`::

        sage: A = matrix(PR, 5, lambda i, j: x^(i + j))
        sage: changer = SDPSparseBasisChanger([ x^k for k in xrange(9) ], [ laguerre(k, x) for k in xrange(9) ])
        sage: cc.add_constraint(A * X == p, changer)
        
    If some polynomial in the constraint cannot be expanded in the
    given basis, then an exception is raised::

        sage: A[2, 2] = x^10
        sage: cc.add_constraint(A * X == p, changer)
        ValueError: Could not expand polynomial in the given basis

    Adding polynomial constraints can be time-consuming. So a progress
    indicator can be shown::

        sage: A = random_matrix(PR, 50)
        sage: X = SDPVariable(50)
        sage: cc.add_constraint(A * X == p, progress = True)
        Adding constraint [========================================] 1275/1275

    A custom message can also be shown::

        sage: cc.add_constraint(A * X == p, progress = True, progress_msg = 'Example')
        Example [========================================] 1275/1275    
    """
    def __init__(self):
        """
        __init__(self)

        Create an empty container.
        """
        self._container = SDPConstraintContainer_wrapper()
        self._var_set = set()
        
    def add_constraint(self, constraint, basis_changer = None, progress = False, progress_msg = None):
        """
        add_constraint(self, constraint, basis_changer = None, progress = False, progress_msg = None)

        Add a constraint to the container.

        INPUT:

        - ``constraint`` -- an instance of SDPLinearConstraint or
          SDPPolynomialConstraint; the constraint to be added.

        - ``basis_changer`` -- if the constraint is a polynomial
          constraint, then a basis changer (an instance of
          SDPSparseBasisChanger) can be specified. This parameter is
          ignored for linear constraints.

        - ``progress`` -- if True, then a progress indicator will be
          shown when adding a polynomial constraint.

        - ``progress_msg`` -- if this is a string, then it specifies
          the message of the progress indicator. ``None`` produces a
          default message.

        EXCEPTIONS:

        - ``ValueError`` -- if the constraint being added is seen to
          be infeasible or if some polynomial in a polynomial
          constraint cannot be expanded on the given basis.

        .. note::

           Environment parameter ``output_eps``
           (cf. :ref:`sdpenvironment`) is used to determine which
           matrix entries are nonzero when processing a
           constraint. Only numbers whose absolute value are greater
           than ``output_eps`` are considered nonzero.
        """
        if isinstance(constraint, SDPLinearConstraint):
            if self._container.add_linear_constraint(constraint):
                raise ValueError('Constraint is infeasible')
        elif isinstance(constraint, SDPPolynomialConstraint):
            ret = self._container.add_poly_constraint(constraint, basis_changer, 
                                                      progress, progress_msg)

            if ret == INFEASIBLE_CONSTRAINT_ERROR:
                raise ValueError('Constraint is infeasible')
            elif ret == REPRESENTATION_ERROR:
                raise ValueError('Could not expand polynomial in the given basis')
            elif ret:
                raise ValueError('Unknown error')
        else:
            raise ValueError('Constraint must be SDPLinearConstraint or SDPPolynomialConstraint')

        for X in constraint.variables():
            self._var_set.add(X)

    def num_constraints(self):
        """
        num_constraints(self)

        Return number of raw constraints in the container.
        """
        return self._container.num_constraints()

    def variables(self):
        """
        variables(self)

        Return a list of variables used in the constraints contained.
        """
        return list(self._var_set)

    def clear(self):
        """
        clear(self)

        Clear the container, erasing all constraints.
        """
        self._container.clear()
        
    def __repr__(self):
        return 'SDPSL constraint container with %d raw constraint(s)' \
          % self.num_constraints()


# Local variables:
# mode: python
# End:
