#
# SDPSL Matrix and MatrixMask
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

# cimport cython
# from libc.stdlib cimport malloc, free
# from sage.matrix.constructor import matrix as sage_matrix
# from sage.matrix.matrix import is_Matrix
# from sage.rings.real_mpfr import is_RealField

# from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing, is_LaurentPolynomialRing
# from sage.rings.polynomial.laurent_polynomial import is_Element as is_LaurentPolynomial
# from sage.rings.polynomial.polynomial_ring import is_PolynomialRing
# from sage.rings.polynomial.polynomial_element import is_Polynomial
# from sage.rings.polynomial.multi_polynomial_ring import is_MPolynomialRing
# from sage.rings.polynomial.multi_polynomial_element import is_MPolynomial

cdef class SDPMatrix_wrapper:
    """
    A wrapper for Matrix<mpf_class> from SDPSL.
    """
    cdef NumberMatrix _matrix
    cdef RealField_class _parent

    def __cinit__(self, RF, arg):
        """
        Creates a new wrapper for an SDPSL matrix of numbers.

        INPUT:

        - ``RF`` -- The RealField that provides the precision of the matrix.

        - ``arg`` -- Either a matrix, that provides the initial data,
          or a triple (m, n, init), where m and n give the number of
          rows and columns of the matrix, and init is the value of
          every entry of the matrix.
        """
        cdef RealNumber tmp

        self._parent = RF
        
        if is_Matrix(arg):
            m = arg.nrows()
            n = arg.ncols()

            try:
                tmp = RealNumber(RF, 0)
                
                sig_on()
                self._matrix = create_number_matrix(m, n, RF.prec(), tmp.value)
                
                for i from 0 <= i < m:
                    for j from 0 <= j < n:
                        tmp = RealNumber(RF, arg[i, j])
                        set_number_entry(self._matrix, i, j, tmp.value)
            finally:
                sig_off()
        else:
            m, n, initial_value = arg

            tmp = RealNumber(RF, initial_value)
            
            try:
                sig_on()
                self._matrix = create_number_matrix(m, n, RF.prec(), tmp.value)
            finally:
                sig_off()

    def __dealloc__(self):
        destroy_number_matrix(self._matrix)

    def get(self, i, j):
        """
        Returns value of entry (i, j) of the matrix.
        """
        cdef RealNumber ret = RealNumber(self._parent, 0)
        get_number_entry(self._matrix, i, j, ret.value)

        return ret
        
    def set(self, i, j, val):
        """
        Sets value of entry (i, j) to val.
        """
        cdef RealNumber num = RealNumber(self._parent, val)
        set_number_entry(self._matrix, i, j, num.value)

    def fill_matrix(self, A, RF):
        """
        Fills a SAGE matrix A with the contents of self, using RF as
        the RealField that gives the precision.
        """
        cdef RealNumber tmp = RealNumber(RF, 0)

        m, n = A.nrows(), A.ncols()

        try:
            sig_on()
            for i from 0 <= i < m:
                for j from 0 <= j < n:
                    get_number_entry(self._matrix, i, j, tmp.value)
                    A[i, j] = RealNumber(RF, tmp)
        finally:
            sig_off()

            
class SDPMatrix:
    """
    An SDPSL numerical matrix.

    SDPSL numerical matrices store arbitrary-precision floating-point
    numbers via GMP.

    EXAMPLES:

    Matrices can be accessed with the ``[]`` operator::

        sage: A = SDPMatrix(5)  # Creates a 5 x 5 matrix
        sage: A[0, 1] = 2
        sage: A[0, 0]
        0.000000000000000
        sage: A[0, 1]
        2.00000000000000
        sage: A[0, 5]
        IndexError: Matrix index out of range

    A precision can be specified for the matrix, in bits; otherwise
    the standard precision is used (see
    :func:`sdpsl_set_gmp_prec`). The initial value of the entries can
    also be specified::

        sage: A = SDPMatrix(5, prec = 128, initial_value = 2)
        sage: A[0, 0]
        2.0000000000000000000000000000000000000

    Matrices can also be initialized to have entries that are equal to
    the entries of a given SAGE matrix over the ``RealField``::

        sage: B = random_matrix(RR, 5)
        sage: A = SDPMatrix(B)
    """
    def __init__(self, arg, initial_value = 0, prec = None):
        """
        __init__(arg, initial_value = 0, prec = None)
        
        Initialize an SDPSL numerical matrix.

        INPUT:
        
        - ``arg`` -- is either an integer giving the number of rows
          and columns of the matrix, a pair ``(m, n)`` of integers
          giving the number of rows and columns of the matrix, or a
          SAGE matrix over the ``RealField`` that will be copied.
          
        - ``prec`` -- is the precision used for the RealField in case
          the dimensions of the matrix are specified; when equal to
          ``None`` precision defaults to the default GMP precision
          used by SDPSL (see :func:`sdpsl_set_gmp_prec`). When a
          SAGE matrix is given, then prec is ignored, and the
          precision of the ``RealField`` of the SAGE matrix is used.

        - ``initial_value`` -- is the initial value of all entries of
          the matrix. When a matrix is given in ``arg``, then this is
          ignored.
        """
        if is_Matrix(arg):
            if not is_RealField(arg.base_ring()):
                raise ValueError('Matrix must be over RealField')

            self._m, self._n = arg.nrows(), arg.ncols()
            self._base_ring = arg.base_ring()
            self._matrix = SDPMatrix_wrapper(self._base_ring, arg)
        else:
            try:
                m, n = arg
            except:
                m = n = arg

            m, n = int(m), int(n)

            if m <= 0 or n <= 0:
                raise ValueError('Dimensions must be positive')

            self._m, self._n = m, n

            if prec is None:
                self._base_ring = RealField(SDPEnvironment.gmp_prec)
            else:
                self._base_ring = RealField(prec)
                
            self._matrix = SDPMatrix_wrapper(self._base_ring, (m, n, self._base_ring(initial_value)))

    def nrows(self):
        """
        nrows(self)
        
        Return the number of rows of self.
        """
        return self._m

    def ncols(self):
        """
        ncols(self)
        
        Return the number of columns of self.
        """
        return self._n

    def is_square(self):
        """
        is_square(self)
        
        Return True iff the matrix is square.
        """
        return self._m == self._n
    
    def __getitem__(self, pos):
        """
        __getitem__(self, pos)
        
        Return the entry at position ``pos``, which is a pair of integers.
        """        
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')

        return self._matrix.get(i, j)

    def __setitem__(self, pos, val):
        """
        __setitem__(self, pos, val)
        
        Set the entry at position ``pos`` to ``val``; ``pos`` is a
        pair of integers.
        """
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')

        self._matrix.set(i, j, self._base_ring(val))

    def sage_matrix(self):
        """
        sage_matrix(self)
        
        Return a SAGE matrix over ``RealField`` containing the data of
        the SDPSL matrix.
        """
        A = sage_matrix(self._base_ring, self._m, self._n)
        self._matrix.fill_matrix(A, self._base_ring)
        
        return A

    def fill_sage_matrix(self, A):
        """
        fill_sage_matrix(self, A)
        
        Fill a SAGE matrix over ``RealField`` with the data of the SDPSL
        matrix.
        """
        if not is_Matrix(A) or not is_RealField(A.base_ring()):
            raise ValueError('Matrix must be a SAGE matrix over RealField')

        self._matrix.fill_matrix(A, A.base_ring())
    
    def __repr__(self):
        if self._m > 10 or self._n > 10:
            return ('%d x %d SDPSL matrix over %s'
                    % (self._m, self._n, repr(self._base_ring)))

        return repr(self.sage_matrix())

    def __str__(self):
        return self.sage_matrix().str()
        

cdef class SDPPolyMatrix_wrapper:
    """
    A wrapper for Matrix<Polynomial> from SDPSL.
    """
    cdef PolyMatrix _matrix

    def __cinit__(self, m, n):
        """
        Creates a matrix of polynomials of size m x n.
        """
        try:
            sig_on()
            self._matrix = create_poly_matrix(m, n)
        finally:
            sig_off()

    def __dealloc__(self):
        destroy_poly_matrix(self._matrix)

    def set_entry(self, i, j, p):
        """
        Sets a given entry of the matrix.
        """
        cdef SDPPolynomial_wrapper wrap = p._p
        set_poly_entry(self._matrix, i, j, wrap._p)

    def get_entry(self, i, j, p):
        """
        Gets a given entry of the matrix, putting the result in p.
        """
        cdef SDPPolynomial_wrapper wrap = p._p
        get_poly_entry(self._matrix, i, j, wrap._p)

    def fill_matrix(self, A):
        """
        Fills a SAGE matrix with the data of self.
        """
        m, n = A.nrows(), A.ncols()
        PR = A.base_ring()
        
        p = SDPPolynomial(PR)
        cdef SDPPolynomial_wrapper wrap = p._p

        try:
            sig_on()
            for i from 0 <= i < m:
                for j from 0 <= j < n:
                    get_poly_entry(self._matrix, i, j, wrap._p)
                    A[i, j] = p.to_polynomial()
        finally:
            sig_off()        

        
class SDPPolyMatrix:
    """
    An SDPSL matrix of polynomials.

    SDPSL polynomial matrices store SDPSL polynomials. SDPSL has its
    own internal representation for Laurent polynomials whose
    coefficients are arbitrary-precision floating-point numbers via
    GMP. (Laurent polynomials are polynomials in which variables can
    have negative exponents.)

    The polynomials stored are transformed to and from SAGE
    polynomials when the matrix is accessed, so the user never sees
    the internal representation of SDPSL.

    EXAMPLES:

    For a matrix to be created, the underlying polynomial ring (which
    must be over ``RealField``) has to be given. This can be a
    polynomial ring or a ring of Laurent polynomials::

        sage: PR = PolynomialRing(RR, 'x')
        sage: A = SDPPolyMatrix((2, 3), PR)
        sage: LR = LaurentPolynomialRing(RealField(128), 'x, y')
        sage: B = SDPPolyMatrix(5, LR)

    Above, ``A`` is a 2-by-3 matrix of polynomials in ``PR``; ``B`` is
    a 5-by-5 matrix of Laurent polynomials in ``LR``. Matrices can be
    accessed using the ``[]`` operator::

        sage: x, y = LR.gens()
        sage: B[2, 3] = x^2 + 2 * x^-2 * y^5
        sage: B[2, 3]
        2.0000000000000000000000000000000000000*x^-2*y^5 + x^2

    Matrices can also be initialized so that their entries are copies
    of the entries of a given SAGE matrix of polynomials. The elements
    of the matrix in question must belong to a ring of polynomials or
    of Laurent polynomials over ``RealField``::

        sage: A = SDPMatrix(random_matrix(PR, 5))
    """
    def __init__(self, *args):
        """
        __init__(self, *args)
        
        Initialize an SDPSL matrix of polynomials.

        INPUT:

        - If a matrix of SAGE polynomials or Laurent polynomials is
          given, then it is copied. The polynomials must belong to a
          ring defined over a ``RealField``.

        - Another possibility is to provide the parameters:

          * ``dimen`` -- either an integer, giving the number of rows
            and columns of the matrix (which is then square), or a
            pair of integers giving the number of rows and columns.

          * ``ring`` -- a ring of polynomials or Laurent polynomials
            over a ``RealField``.
        """
        if len(args) == 0:
            raise ValueError('At least 1 argument is expected')

        if is_Matrix(args[0]):
            m, n = args[0].nrows(), args[0].ncols()
            ring = args[0].base_ring()
            
            if is_LaurentPolynomialRing(ring) or is_PolynomialRing(ring) \
              or is_MPolynomialRing(ring):
                if not is_RealField(ring.base_ring()):
                    raise ValueError('Polynomial ring must be over RealField')

                self._m, self._n = m, n
                self._parent = ring
                self._base_ring = ring.base_ring()
                self._matrix = SDPPolyMatrix_wrapper(m, n)

                try:
                    sig_on()
                    for i from 0 <= i < m:
                        for j from 0 <= j < n:
                            self._matrix.set_entry(i, j, SDPPolynomial(args[0][i, j]))
                finally:
                    sig_off()
            else:
                raise ValueError('Matrix must be over a real polynomial ring')
        else:
            try:
                m, n = args[0]
            except:
                m = n = args[0]

            m, n = int(m), int(n)

            self._m, self._n = m, n

            if len(args) != 2:
                raise ValueError('Expected dimensions and a real polynomial ring')

            ring = args[1]

            if is_LaurentPolynomialRing(ring) or is_PolynomialRing(ring) \
              or is_MPolynomialRing(ring):
              if not is_RealField(ring.base_ring()):
                  raise ValueError('Polynomial ring must be over RealField')

              self._parent = ring
              self._base_ring = ring.base_ring()
            else:
                raise ValueError('A polynomial ring is required')

            self._matrix = SDPPolyMatrix_wrapper(m, n)

    def nrows(self):
        """
        nrows(self)
        
        Return the number of rows in the matrix.
        """
        return self._m

    def ncols(self):
        """
        ncols(self)
        
        Return the number of columns in the matrix.
        """
        return self._n
            
    def is_square(self):
        """
        is_square(self)
        
        Return True iff the matrix is square.
        """
        return self._m == self._n

    def __getitem__(self, pos):
        """
        __getitem__(self, pos)
        
        Return the entry at position ``pos``, which is a pair of integers.
        """                
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')
        
        tmp = SDPPolynomial(self._parent)
        self._matrix.get_entry(i, j, tmp)

        return tmp.to_polynomial()

    def __setitem__(self, pos, val):
        """
        __setitem__(self, pos, val)
        
        Set the entry at position ``pos`` to ``val``; ``pos`` is a
        pair of integers.
        """
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')

        tmp = SDPPolynomial(self._parent(val))
        self._matrix.set_entry(i, j, tmp)

    def sage_matrix(self):
        """
        sage_matrix(self)
        
        Return the corresponding SAGE matrix.
        """
        A = sage_matrix(self._parent, self._m, self._n)
        self._matrix.fill_matrix(A)
        
        return A

    def __repr__(self):
        if self._m > 10 or self._n > 10:
            return ('%d x %d SDPSL matrix over %s'
                    % (self._m, self._n, repr(self._parent)))

        return repr(self.sage_matrix())

    def __str__(self):
        return self.sage_matrix().str()

    
def is_poly_matrix(A):
    """
    Returns True iff A is an instance of SDPPolyMatrix or is a matrix
    of polynomials over a real polynomial ring or a real Laurent
    polynomial ring.
    """
    if isinstance(A, SDPPolyMatrix):
        return True

    if not is_Matrix(A):
        return False

    ring = A.base_ring()

    if is_PolynomialRing(ring) or is_MPolynomialRing(ring) \
      or is_LaurentPolynomialRing(ring):
      F = ring.base_ring()

      if is_RealField(F):
          return True
      else:
          return False

    return False
      

cdef class SDPMatrixMask_wrapper:
    """
    Wrapper for MatrixMask from SDPSL.
    """
    cdef SDPSLMatrixMask _mask

    def __cinit__(self, m, n):
        """
        Creates an m x n matrix mask.
        """
        self._mask = create_matrix_mask(m, n)

    def __dealloc__(self):
        destroy_matrix_mask(self._mask)

    def is_on(self, i, j):
        return bool(matrix_mask_is_on(self._mask, i, j))

    def turn_on(self, i, j):
        matrix_mask_turn_on(self._mask, i, j)

    def turn_off(self, i, j):
        matrix_mask_turn_off(self._mask, i, j)
        
    def clear(self):
        clear_matrix_mask(self._mask)
        

class SDPMatrixMask:
    """
    An SDPSL MatrixMask.

    A matrix mask is a matrix of bits that keeps a list of all bits
    that are on for easy access. Checking whether a bit is on or off
    takes constant time, and iterating over the list of bits turned on
    takes constant time per bit that is turned on.

    Such masks can be used to specify which entries of a matrix are to
    be considered nonzeros. Such sparsity information can be exploited
    by some functions to achieve higher efficiency.

    EXAMPLES:

    Masks can be created given their dimensions::

        sage: A = SDPMatrixMask(5)  # a 5 x 5 mask
        sage: B = SDPMatrixMask((2, 3))  # a 2 x 3 mask

    Masks can be accessed using the ``[]`` operator::

        sage: A[0, 2] = True
        sage: A
        00100
        00000
        00000
        00000
        00000
        sage: A[0, 2] = False
        sage: A[0, 2]
        False
    """
    def __init__(self, dimen):
        """
        __init__(self, dimen)
        
        Initialize an SDPSL matrix mask.
        
        INPUT:

        - ``dimen`` -- either an integer, giving the number of rows
          and columns of a square mask, or a pair of integers giving
          the number of rows and columns.
        """
        try:
            m, n = dimen
        except:
            m = n = dimen

        m, n = int(m), int(n)

        if m <= 0 or n <= 0:
            raise ValueError('Dimensions must be positive')

        self._m, self._n = m, n
        self._mask = SDPMatrixMask_wrapper(m, n)
        self._on_list = [ ]

    def nrows(self):
        """
        nrows(self)
        
        Return number of rows of the mask.
        """
        return self._m

    def ncols(self):
        """
        ncols(self)
        
        Return number of columns of the mask.
        """
        return self._n
        
    def __getitem__(self, pos):
        """
        __getitem__(self, pos)
        
        Return the entry at position ``pos``, which is a pair of integers.
        """                        
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')

        return self._mask.is_on(i, j)

    def __setitem__(self, pos, val):
        """
        __setitem__(self, pos, val)
        
        Set the entry at position ``pos`` to ``val``; ``pos`` is a
        pair of integers. It must be possible to convert ``val`` to a
        boolean.
        """
        i, j = pos
        i, j = int(i), int(j)

        if i < 0 or i >= self._m or j < 0 or j >= self._n:
            raise IndexError('Matrix index out of range')

        if val == True:
            if not self._mask.is_on(i, j):
                self._on_list.append((i, j))
                
            self._mask.turn_on(i, j)
        else:
            if self._mask.is_on(i, j):
                self._on_list.remove((i, j))
                
            self._mask.turn_off(i, j)

    def clear(self):
        """
        clear(self)
        
        Clear the mask, setting all bits to False.
        """
        self._mask.clear()
        self._on_list = [ ]

    def on_list(self):
        """
        on_list(self)

        Return the list of positions turned on.
        """
        return list(self._on_list)
        
    def __repr__(self):
        if self._m < 25 and self._n < 50:
            return self.__str__()
        
        return '%d x %d SDPSL MatrixMask' % (self._m, self._n)

    def __str__(self):
        lines = [ ]

        for i from 0 <= i < self._m:
            s = ''
            for j from 0 <= j < self._n:
                if self._mask.is_on(i, j):
                    s += '1'
                else:
                    s += '0'

            lines.append(s)

        return '\n'.join(lines)

    
# Local variables:
# mode: python
# End:
