#
# SDPSL Polynomial
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

# cimport cython
# from libc.stdlib cimport malloc, free
# from sage.rings.real_mpfr cimport RealNumber, RealField, RealField_class
# from sage.rings.real_mpfr import is_RealField
# from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing, is_LaurentPolynomialRing
# from sage.rings.polynomial.laurent_polynomial import is_Element as is_LaurentPolynomial
# from sage.rings.polynomial.polynomial_ring import is_PolynomialRing
# from sage.rings.polynomial.polynomial_element import is_Polynomial
# from sage.rings.polynomial.multi_polynomial_ring import is_MPolynomialRing
# from sage.rings.polynomial.multi_polynomial_element import is_MPolynomial

def is_polynomial(p):
    """
    Returns True iff p is a polynomial or a Laurent polynomial.
    """
    if is_Polynomial(p) or is_MPolynomial(p):
            return True

    if is_LaurentPolynomial(p):
        try:
            ring = p.parent()
        except:
            return False

        if is_LaurentPolynomialRing(ring):
            return True

    return False

def is_real_polynomial(p):
    """
    Returns True iff p is a polynomial or a Laurent polynomial over RealField.
    """
    if is_Polynomial(p) or is_MPolynomial(p):
        if is_RealField(p.base_ring()):
            return True

    if is_LaurentPolynomial(p):
        try:
            ring = p.parent()
        except:
            return False

        if is_LaurentPolynomialRing(ring) and is_RealField(p.base_ring()):
            return True

    return False


def make_real_poly(p):
    """
    Tries to make a real polynomial out of ``p``.
    """
    if not is_real_polynomial(p):
        if is_Polynomial(p) or is_MPolynomial(p):
            ring = p.parent()
            new_ring = PolynomialRing(RealField(SDPEnvironment.gmp_prec), ','.join(ring.variable_names()))
        elif is_LaurentPolynomial(p):
            try:
                if is_LaurentPolynomialRing(p.parent()): 
                    # is_LaurentPolynomial says that pi, for instance, is a Laurent polynomial
                    # So we check also the parent ring
                    ring = p.parent()
                    new_ring = lambda q: LaurentPolynomialRing(RealField(SDPEnvironment.gmp_prec), ','.join(ring.variable_names()))(q.dict())
                else:
                    new_ring = PolynomialRing(RealField(SDPEnvironment.gmp_prec), 'x')
            except:
                new_ring = PolynomialRing(RealField(SDPEnvironment.gmp_prec), 'x')
        else:
            new_ring = PolynomialRing(RealField(SDPEnvironment.gmp_prec), 'x')

        try:
            p = new_ring(p)
        except:
            raise ValueError('Could not convert polynomial to a real polynomial')

    return p


def add_poly_variables(P):
    """
    Given a polynomial ring P, adds the variables of the ring to
    SDPSL, if the ring has more variables than are defined in SDPSL.
    """
    var_names = P.variable_names()
    n = get_num_variables()
    
    if len(var_names) > n:
        for name in var_names[n:]:
            add_poly_variable(name)

            
cdef class SDPPolynomial_wrapper:
    """
    A wrapper for Polynomial from SDPSL.
    """
    cdef SDPSLPoly _p

    def __cinit__(self, p = None):
        if p is None:
            self._p = create_polynomial()
        else:
            self._p = <SDPSLPoly> p

    def __dealloc__(self):
        destroy_polynomial(self._p)

    def set_polynomial(self, other, RF):
        """
        This sets the polynomial to a given SAGE polynomial, using RF
        as the RealField that specifies the precision to be used.
        This function should be called only once, or the result is
        just a mess. BIG FIXME!
        """
        add_poly_variables(other.parent())

        coeffs = other.coefficients()
        exponents = other.exponents()

        if len(coeffs) == 0:
            return

        # Sometimes the exponents are a tuple; for univariate
        # polynomials they are just numbers. This we need to check.
        is_tuple = True
        try:
            foo = iter(exponents[0])
        except:
            is_tuple = False
        
        num_coeffs = len(coeffs)
        
        cdef int n = other.parent().ngens()
        cdef int *exps = <int *> malloc(n * cython.sizeof(int))

        if exps is NULL:
            raise MemoryError('Cannot allocate memory for exps')

        # FIXME: We create new numbers at every iteration? It seems
        # this could be better done.
        cdef RealNumber tmp
        cdef int prec = RF.prec()

        if not is_tuple:
            for k from 0 <= k < num_coeffs:
                exps[0] = int(exponents[k])

                tmp = RealNumber(RF, coeffs[k])
                add_monomial(self._p, n, exps, tmp.value, prec)
        else:
            for k from 0 <= k < num_coeffs:
                for i from 0 <= i < n:
                    exps[i] = int(exponents[k][i])

                tmp = RealNumber(RF, coeffs[k])
                add_monomial(self._p, n, exps, tmp.value, prec)

        free(exps)

    def to_polynomial(self, PR, RF):
        """
        Returns the SAGE polynomial, element of polynomial ring PR,
        corresponding to self.
        """
        # First we interface with SDPSL
        cdef int n, ncoeffs
        cdef int *exponents
        cdef MPFRNumber *coefficients

        polynomial_to_array(self._p, cython.address(n), cython.address(ncoeffs), 
                            cython.address(exponents), cython.address(coefficients), RF.prec())

        # Then we assemble a SAGE Laurent polynomial or polynomial
        cdef RealNumber tmp = RealNumber(RF, 0)
        num_vars = PR.ngens()
        univariate = is_PolynomialRing(PR)

        mons = { }
        for k from 0 <= k < ncoeffs:
            if univariate:
                t = exponents[k * n]
            else:
                t = tuple()
                for i from 0 <= i < num_vars:
                    t += (exponents[k * n + i],)
                
            mpfr_set(tmp.value, (<mpfr_t *> coefficients)[k], <mpfr_rnd_t> 0)
                
            mons[t] = RealNumber(RF, tmp)

        # Then we free memory
        free(exponents)
        free(coefficients)

        if mons:
            return PR(mons)
        else:
            return PR(0)
        
class SDPPolynomial:
    """
    An SDPSL polynomial. 
    """
    
    def __init__(self, init_obj):
        """
        Initializes a new polynomial.

        INPUT:

        - ``init_obj`` -- this can be either a polynomial or Laurent
          polynomial over the RealField, and then it is copied, or a
          polynomial ring or a Laurent polynomial ring over the
          RealField, in which case the polynomial is initialized to
          zero.
        """
        if is_LaurentPolynomialRing(init_obj) or is_PolynomialRing(init_obj) \
            or is_MPolynomialRing(init_obj):
            if not is_RealField(init_obj.base_ring()):
                raise ValueError('Ring must be over RealField')

            self._parent = init_obj
            self._base_ring = init_obj.base_ring()

            self._p = SDPPolynomial_wrapper()
        elif is_real_polynomial(init_obj):
            self._parent = init_obj.parent()
            self._base_ring = init_obj.base_ring()
            
            self._p = SDPPolynomial_wrapper()
            self._p.set_polynomial(init_obj, self._base_ring)
        else:
            raise ValueError('Expected either real polynomial ring or real polynomial')

    def to_polynomial(self):
        """
        Returns the SAGE polynomial corresponding to self.
        """
        return self._p.to_polynomial(self._parent, self._base_ring)

    def __repr__(self):
        return repr(self._p.to_polynomial(self._parent, self._base_ring))
            

# Local variables:
# mode: python
# End:
