#
# SDPSL SDPProblem
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

cdef class SDPProblem_wrapper:
    """
    Wrapper for SDPProblem from SDPSL.
    """
    cdef SDPSLProblem _problem

    def __cinit__(self):
        self._problem = create_problem()

    def __dealloc__(self):
        destroy_problem(self._problem)

    def set_objective(self, C):
        cdef SDPSLLinearCombination tmp = make_linear_combination(C)

        try:
            sig_on()
            problem_set_objective(self._problem, tmp)
        finally:
            destroy_linear_combination(tmp)
            sig_off()

    def add_container(self, cc):
        cdef SDPConstraintContainer_wrapper tmp = cc._container
        problem_add_container(self._problem, tmp._container)

    def write_sdpa(self, filename, prec):
        try:
            sig_on()
            ret = write_sdpa(self._problem, filename, prec)
        finally:
            sig_off()

        return ret

    
class SDPProblem:
    """
    An SDPSL SDP problem.

    A problem contains an objective function and a list of constraint
    containers that provide its constraints.

    EXAMPLES::

        sage: X = SDPVariable(5)
        sage: A = random_matrix(RR, 5)
        sage: B = random_matrix(RR, 5)
        sage: cc = SDPConstraintContainer()
        sage: cc.add_constraint(A * X == 2)
        sage: problem = SDPProblem()
        sage: problem.set_objective(B * X)
        sage: problem.add_container(cc)
        sage: problem
        SDPSL problem with 1 constraint container(s) totaling 1 raw constraint(s)

    An empty objective function means that we have a pure feasibility
    problem::

        sage: problem.set_objective(SDPLinearCombination())
        sage: problem
        SDPSL pure feasibility problem with 1 constraint container(s) totaling 1 raw constraint(s)

    Problems can be written to files in sparse SDPA format::

        sage: problem.write_sdpa('out.sdpa')
    """
    def __init__(self):
        """
        __init__(self)

        Create an empty problem.
        """
        self._problem = SDPProblem_wrapper()
        self._containers = [ ]
        self._obj_variables = [ ]

    def set_objective(self, C):
        """
        set_objective(self, C)

        Set the objective function of the problem.

        INPUT:

        - ``C`` -- is an :class:`SDPLinearCombination` containing the objective
          function; the data of the combination is copied and it is no
          longer needed afterwards.
        """
        if not isinstance(C, SDPLinearCombination):
            raise ValueError('Objective function must be instance of SDPLinearCombination')
        
        self._problem.set_objective(C)
        self._obj_variables = C.variables()

    def add_container(self, cc):
        """
        add_container(self, cc)

        Add constraint container ``cc`` to the problem.
        """
        if not isinstance(cc, SDPConstraintContainer):
            raise ValueError('Container must be instance of SDPConstraintContainer')
        
        self._problem.add_container(cc)
        self._containers.append(cc)

    def containers(self):
        """
        containers(self)

        Return list of constraint containers in the problem.
        """
        return list(self._containers)

    def variables(self):
        """
        variables(self)

        Return list of variables occurring in the problem.
        """
        variables = list(self._obj_variables)

        for cc in self._containers:
            variables += cc.variables()

        return list(set(variables))

    def write_sdpa(self, filename, prec = 15):
        """
        write_sdpa(self, filename, prec = 15)

        Write the problem to a file in sparse SDPA format.

        INPUT:

        - ``filename`` -- is the name of the output file.

        - ``prec`` -- how many digits of precision to use for the
          output.
        """
        ret = self._problem.write_sdpa(filename, prec)

        if ret == FILE_ERROR:
            raise IOError('Cannot write to file "%s"' % filename)
        elif ret == VALUE_ERROR:
            raise ValueError('Problem is empty')

    def __repr__(self):
        feas = ''
        if not self._obj_variables:
            feas = 'pure feasibility '
            
        return 'SDPSL %sproblem with %d constraint container(s) totaling %d raw constraint(s)' \
          % (feas, len(self._containers), sum(cc.num_constraints() for cc in self._containers))
        

# Local variables:
# mode: python
# End:

