#
# SDPSL SDPVariable
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

cdef class SDPVariable_wrapper:
    """
    A wrapper for SDPVariable from SDPSL.
    """
    cdef SDPSLVariable _var

    def __cinit__(self, size):
        self._var = create_variable(size)

    def __dealloc__(self):
        destroy_variable(self._var)

    def idx(self):
        return get_variable_idx(self._var)
    

class SDPVariable:
    """
    A variable in an SDPSL problem.
    """
    def __init__(self, size):
        """
        __init__(self, size)
        
        Create a new variable.

        Variables are uniquely identified by an index, which is
        assigned to them at creation. Two variables are considered
        equal if they have the same index.

        INPUT:

        - ``size`` -- the size of the variable matrix.

        Variables can be multiplied on the left by matrices or pairs
        of a matrix and a mask to give instances of
        :class:`SDPLinearCombination` or
        :class:`SDPPolynomialCombination`. See the documentation of
        these classes for more information.

        EXAMPLES::

            sage: from sage.sdpsl import *
            sage: SDPVariable(5)
            SDPSL variable of size 5 with index 0
        """
        self._size = int(size)
        
        if self._size <= 0:
            raise ValueError('Size must be positive')

        self._var = SDPVariable_wrapper(self._size)
        self._idx = self._var.idx()

    def idx(self):
        """
        idx(self)
        
        Return the index of the variable.
        """
        return self._idx

    def size(self):
        """
        size(self)
        
        Return the variable size.
        """
        return self._size

    def __hash__(self):
        return hash(self._idx)

    def __eq__(self, other):
        if not isinstance(other, SDPVariable):
            return NotImplemented
        
        return self._idx == other._idx

    def __rmul__(self, other):
        if type(other) is tuple:
            try:
                A, mask = other
            except:
                A = other
        else:
            A = other

        if isinstance(A, SDPPolyMatrix):
            return SDPPolynomialCombination([ (self, other) ])

        if is_Matrix(A):
            ring = A.base_ring()

            if is_PolynomialRing(ring) or is_MPolynomialRing(ring) \
              or is_LaurentPolynomialRing(ring):
                return SDPPolynomialCombination([ (self, other) ])
            else:
                return SDPLinearCombination([ (self, other) ])
        else:
            return SDPLinearCombination([ (self, other) ])
    
    def __repr__(self):
        return 'SDPSL variable of size %d with index %d' % (self._size, self._idx)


# Local variables:
# mode: python
# End:
