/*
  progbar.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/progbar.h>
#include <iomanip>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*
  Constructor
*/
ProgressBar::ProgressBar(const std::string& message, int bar_size, int max_value,
			 int update_step, char bar_char, std::ostream& out)
  : _bar_char(bar_char), message_text(message), _bar_size(bar_size),
    _update_step(update_step), _out(out)
{
  set_max(max_value);

  /* The initial values */
  cur = show_cur = 0;
}

/*
  set_max
*/
void ProgressBar::set_max(int max_value)
{
  _max_value = max_value;

  unit_value = (double) _max_value / _bar_size;

  /* Set default value for update_step if needed */
  if (_update_step == 0) 
    cur_update_step = unit_value;
  else
    cur_update_step = _update_step;

  /* Computes counter_wd (we assume max > 0) */
  counter_wd = 0;
  for (int tmp = _max_value; tmp > 0; tmp /= 10) counter_wd++;
}

/*
  start
*/
void ProgressBar::start()
{
  cur = show_cur = 0;
  update();
}

/*
  finish
*/
void ProgressBar::finish()
{
  cur = _max_value;
  update();
  _out << endl;
}

/*
  operator++
*/
ProgressBar& ProgressBar::operator++()
{
  update_mutex.lock();

  cur++;
  update();

  update_mutex.unlock();

  return *this;
}

/*
  operator +=
*/
ProgressBar& ProgressBar::operator+=(int val)
{
  update_mutex.lock();

  cur += val;
  update();

  update_mutex.unlock();

  return *this;
}

/*
  update
*/
void ProgressBar::update()
{
  if (cur == 0 || cur - show_cur >= cur_update_step || cur == _max_value) {
    /* Displays the message */
    _out << "\r" << message_text << " [";

    /* Displays appropriate number of bar characters */
    int c = 0;
    for (c = 0; c < cur / unit_value && c < _bar_size; c++) 
      _out << _bar_char;

    for ( ; c < _bar_size; c++)
      _out << " ";

    /* Shows counter */
    _out << "] ";
    _out << setw(counter_wd) << cur << "/" << _max_value;
    _out.flush();

    /* And finishes the update */
    show_cur = cur;
  }
}

}  // namespace
