/*
  sparsechanger.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/sparsechanger.h>
#include <cassert>
#include <sstream>
#include <sdpsl/linsys.h>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*
  Constructors
*/
SparseBasisChanger::SparseBasisChanger(const SparseMonomialBasis& mon_basis,
				       const std::vector<Polynomial>& poly_basis,
				       const SDPEnvironment& e)
  : env(e), size(mon_basis.basis_size()), _mon_basis(mon_basis), 
    Ainv(mon_basis.basis_size(), mon_basis.basis_size(), 0)
{
  if (mon_basis.basis_size() != poly_basis.size())
    throw InvalidBasis(InvalidBasis::incompatible_bases, ERROR_MSG("The bases provided are incompatible"));
  
  /* We fill the basis change matrix */
  Matrix<Number> A(size, size, 0);

  for (int j = 0; j < size; j++)
    for (auto& c : poly_basis[j]) {
      int i = _mon_basis.mon_idx(c.first);

      if (i < 0)
	throw InvalidBasis(InvalidBasis::incompatible_bases, ERROR_MSG("The bases provided are incompatible"));

      A[i][j] = c.second;
    }

  /* Initializes the solver */
  typedef std::unique_ptr<LinearSystemSolver<Number, NumberOperations> > solver_ptr;
  solver_ptr solver;

  try {
    solver = solver_ptr(new LinearSystemSolver<Number, NumberOperations>(A, env.tolerance_eps()));
  }
  catch (LinearSystemSolver<Number, NumberOperations>::SingularMatrix) {
    stringstream ss;
    ss << "Basis change matrix is singular with tolerance_eps = " << env.tolerance_eps();

    throw InvalidBasis(InvalidBasis::singular_basis_change_matrix, ERROR_MSG(ss.str()));
  }
  
  /* Fill in the inverse matrix */
  vector<Number> mon_vec(size, 0);
  vector<Number> coeffs(size, 0);

  row_masks.reserve(size);

  for (int i = 0; i < size; i++) {
    mon_vec[i] = 1;
    solver->solve_for(mon_vec, coeffs);
    mon_vec[i] = 0;

    row_masks.emplace_back(size);
    
    for (int j = 0; j < size; j++) {
      Ainv[i][j] = coeffs[j];

      if (NumberOperations::abs(coeffs[j]) > env.output_eps())
	row_masks[i].turn_on(j);
    }
  }
}

/*
  basis_coefficients
*/
bool SparseBasisChanger::basis_coefficients(const Polynomial& p, std::vector<Number>& coeffs,
					    VectorMask& mask) const
{
  mask.clear();

  for (auto& c : p) {
    int i = _mon_basis.mon_idx(c.first);

    if (i < 0) return false;

    for (int j : row_masks[i])
      if (mask.is_on(j))
	coeffs[j] += c.second * Ainv[i][j];
      else {
	coeffs[j] = c.second * Ainv[i][j];
	mask.turn_on(j);
      }
  }

  return true;
}


}  // namespace
