/*
  sparsemon.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/sparsemon.h>
#include <sdpsl/sdperror.h>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*
  Constructors
*/
SparseMonomialBasis::SparseMonomialBasis(const std::vector<Monomial>& basis)
{
  for (int i = 0; i < basis.size(); i++) {
    auto ret = basis_pos.insert(BasisPosMap::value_type(basis[i], i));

    if (!ret.second)
      throw InvalidBasis(InvalidBasis::repeated_monomial, ERROR_INFO);
  }

  _monomials = basis;
}


SparseMonomialBasis::SparseMonomialBasis(const SDPPolynomialConstraint& C)
{
  int k = 0;
  for (auto& s : C.lhs()) {
    const Matrix<Polynomial>& A = s.second.matrix();
    const MatrixMask *mask = s.second.mask();

    if (!mask) 
      for (int i = 0; i < A.num_rows(); i++)
	for (int j = 0; j < A.num_cols(); j++)
	  for (auto& c : A[i][j]) {
	    auto ret = basis_pos.insert(BasisPosMap::value_type(c.first, k));
	    
	    if (ret.second) {
	      k++;
	      _monomials.push_back(c.first);
	    }
	  }
    else
      for (auto& p : *mask)
	for (auto& c : A[p.i][p.j]) {
	  auto ret = basis_pos.insert(BasisPosMap::value_type(c.first, k));
	  
	  if (ret.second) {
	    k++;
	    _monomials.push_back(c.first);
	  }
	}
  }
  
  for (auto& c : C.rhs()) {
    auto ret = basis_pos.insert(BasisPosMap::value_type(c.first, k));
    
    if (ret.second) {
      k++;
      _monomials.push_back(c.first);
    }
  }
}

/*
  basis_coefficients
*/
bool SparseMonomialBasis::basis_coefficients(const Polynomial& p, vector<Number>& coeffs, 
					     VectorMask& mask) const
{
  mask.clear();
  
  for (auto& c : p) {
    auto it = basis_pos.find(c.first);
    
    if (it == basis_pos.end())
      return false;
    
    mask.turn_on(it->second);
    coeffs[it->second] = c.second;
  }
  
  return true;		
}
  
/*
  add_monomial
*/
void SparseMonomialBasis::add_monomial(const Monomial& mon)
{
  auto ret = basis_pos.insert(BasisPosMap::value_type(mon, basis_pos.size()));
 
  if (!ret.second)
    throw InvalidBasis(InvalidBasis::repeated_monomial, ERROR_MSG("Repeated monomial"));

  _monomials.push_back(mon);
}

/*
  mon_idx
*/
int SparseMonomialBasis::mon_idx(const Monomial& mon) const
{
  auto it = basis_pos.find(mon);

  if (it == basis_pos.end())
    return -1;

  return it->second;
}

}  // namespace
