/*!
  \file sdplin.h

  \brief Classes for a linear combination of variables and matrices.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPLIN_H__
#define __SDPLIN_H__

#include <map>
#include <memory>
#include <utility>
#include <sdpsl/number.h>
#include <sdpsl/matrix.h>
#include <sdpsl/matrixmask.h>
#include <sdpsl/sdperror.h>
#include <sdpsl/sdpvar.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A pair of a matrix shared reference and a mask for the matrix.

  Some functions might exploit the fact that matrices are sparse. So
  we use the concept of <i>masked matrices</i>: matrices together
  with a mask that tells us where the nonzeros are.

  This class keeps a \c const shared reference to a matrix and also a
  shared reference to a mask for it. The mask is optional; if it is
  not provided then all entries of the matrix have to be considered
  potential nonzeros.
*/
class SDPMaskedMatrix {
  const_shared_ref<Matrix<Number> > Aref;
  std::unique_ptr<const_shared_ref<MatrixMask> > _mask;

public:
  //! Creates a masked matrix with no mask.
  SDPMaskedMatrix(const Matrix<Number>& mat) : Aref(mat) { }
  
  //! Copy constructor.
  SDPMaskedMatrix(const SDPMaskedMatrix& other)
    : Aref(other.Aref)
  {
    if (other._mask)
      _mask.reset(new const_shared_ref<MatrixMask>(*other._mask));
  }

  /*!
    \brief Creates a masked matrix with the given mask.

    \throws ValueError if the size of the mask differs from the size of the matrix.
  */
  SDPMaskedMatrix(const Matrix<Number>& mat, const MatrixMask& mask)
    : Aref(mat)
  { 
    if (mask.num_rows() != mat.num_rows() || mask.num_cols() != mat.num_cols())
      throw ValueError(ERROR_MSG("Matrix and mask of different sizes"));

    _mask.reset(new const_shared_ref<MatrixMask>(mask));
  }

  //! Returns a reference to the matrix contained.
  const Matrix<Number>& matrix() const { return *Aref; }

  //! Returns a pointer to the mask, which is \c nullptr if there is no mask.
  const MatrixMask *mask() const 
  { 
    if (_mask) return _mask->ptr();
    return nullptr;
  }

  //! Assignment operator is deleted.
  SDPMaskedMatrix& operator=(const SDPMaskedMatrix& other) = delete;
};


/*!
  \brief A map associating numerical matrices with variables for an SDP.

  The left-hand side of a linear constraint of the semidefinite
  programming problems that we consider  (see \ref tut_sdp for details),
  and also the objective function of such a problem, is an expression like

  \f[\sum_{j=1}^r \langle A_j, X_j \rangle,\f]

  where the \f$A_j\f$ are numerical matrices of the appropriate
  sizes.

  Class \c SDPLinearCombination represents expressions of this
  sort. It is a map associating instances of \c SDPMaskedMatrix
  to variables of the problem.

  \section sdplin_iter Iteration

  Iteration is done over pairs <tt>std::pair<SDPVariable,
  SDPMaskedMatrix></tt>. Only constant iteration is allowed. The pairs
  are iterated over in the order imposed by the variable indices.

  \section sdplin_op Constructing linear combinations

  A \ref lin_mult "multiplication operator" is defined for multiplying
  instances of \c SDPMaskedMatrix with instances of \c
  SDPVariable. Such a multiplication yields an instance of \c
  SDPLinearCombination associating the given matrix with the given
  variable.

  Since \c SDPMaskedMatrix has a constructor accepting a reference to
  a \c Matrix<Number>, this means also that multiplication of an
  instance of \c Matrix<Number> with an instance of \c SDPVariable is
  defined. In this case no mask is used.

  Addition of two instances of \c SDPLinearCombination is also defined
  (see \ref lin_add "below"). Together with the multiplication explained
  above, this means that instances of \c SDPLinearCombination can be
  constructed using a very clean multiplication/addition syntax, as
  illustrated in this example:

  \include sdplinadd.cpp

  \subsection sdplin_complex Complexity

  Notice that since \c SDPMaskedMatrix keeps references, no data from
  the matrices or masks is ever copied. In any case, the way in which
  combinations are built might not be efficient when
  many variables are concerned (though \c SDPLinearCombination is
  internally implemented with \c std::map). In semidefinite
  programming problems, however, only a few (maybe a dozen or two)
  variables are used, and therefore this overhead is hardly going to
  affect the running time of the program as a whole. It is to be
  expected that the generation of the constraint matrices will take
  most of the time in any case.

  Moreover, since \c SDPLinearCombination keeps references, if many
  constraints share the same general shape, one may create a single
  instance of \c SDPLinearCombination and, by changing the matrices,
  change the content of the combination itself. If the combination is
  immediately used (for instance, to add a constraint), then it is
  possible to avoid creating many instances of \c SDPLinearCombination.
*/
class SDPLinearCombination {
  typedef std::map<SDPVariable, SDPMaskedMatrix> SummandMap;

  SummandMap summands;

public:
  //! Constructs an empty linear combination.
  SDPLinearCombination() { }

  /*!
    \brief Constructs a linear combination assigning the masked matrix
    \a mm to variable \a X

    \throws ValueError if the matrix is not square or if its size
    differs from that of the variable.
  */
  SDPLinearCombination(const SDPVariable& X, const SDPMaskedMatrix& mm);
  
  //! Constant iterator.
  typedef SummandMap::const_iterator iterator;

  //! Returns iterator to the beginning of the map.
  iterator begin() const { return summands.begin(); }

  //! Returns past-the-end iterator.
  iterator end() const { return summands.end(); }

  /*!
    \brief In-place addition operator.

    Adds the linear combination \a other to this one. We assume that the
    variables occurring in this combination are distinct than those
    occuring in \a other. If this is not the case, an exception is
    thrown.

    \throws ValueError if a variable occurs both in this linear
    combination and \a other. If the exception is thrown then the sum
    is left half-finished.

    \returns a reference to this linear combination.
  */
  SDPLinearCombination& operator+=(const SDPLinearCombination& other);

  friend SDPLinearCombination operator+(const SDPLinearCombination& a, const SDPLinearCombination& b);
};

//
// Prototypes
//

/*! \relates SDPLinearCombination
  \anchor lin_mult

  \brief Multiplication of a masked matrix and a variable.

  Multiplying an instance of \c SDPMaskedMatrix by an instance of \c
  SDPVariable yields an instance of \c SDPLinearCombination
  associating the given masked matrix with the given variable.

  Since \c SDPMaskedMatrix has a constructor accepting a reference to
  \c Matrix<Number>, \cpp will promote any instance of \c
  Matrix<Number> multiplied by an instance of \c SDPVariable to an
  instance of \c SDPMaskedMatrix in order to make the multiplication
  work. This means that one can also multiply instances of \c
  Matrix<Number> with instances of \c SDPVariable.

  \throws ValueError if the matrix is not square or if its size
  differs from that of the variable.
*/
extern SDPLinearCombination operator*(const SDPMaskedMatrix& mat, const SDPVariable& X);

/*! \anchor lin_add

  \brief Adds two linear combinations.

  \throws ValueError if the combinations have variables in common.
*/
extern SDPLinearCombination operator+(const SDPLinearCombination& a, const SDPLinearCombination& b);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
