/*!
  \file sdppol.h

  \brief Classes for linear combinations of polynomial matrices and
  variables.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPPOL_H__
#define __SDPPOL_H__

#include <map>
#include <memory>
#include <utility>
#include <sdpsl/polynomial.h>
#include <sdpsl/matrix.h>
#include <sdpsl/matrixmask.h>
#include <sdpsl/sdpvar.h>
#include <sdpsl/sdperror.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A matrix of polynomials together with a mask.

  Some functions might exploit the fact that a matrix of polynomials
  is sparse, and therefore we allow a mask to be specified which tells
  us which are the nonzero entries of the matrix.

  This class keeps a const shared reference (cf. \c Matrix) for the
  matrix and the mask.
*/
class SDPMaskedPolyMatrix {
  const_shared_ref<Matrix<Polynomial> > Aref;
  std::unique_ptr<const_shared_ref<MatrixMask> > _mask;

public:
  //! Constructs a masked polynomial matrix without a mask.
  SDPMaskedPolyMatrix(const Matrix<Polynomial>& mat) : Aref(mat) { }

  //! Copy constructor.
  SDPMaskedPolyMatrix(const SDPMaskedPolyMatrix& other)
    : Aref(other.Aref) 
  {
    if (other._mask)
      _mask.reset(new const_shared_ref<MatrixMask>(*other._mask));
  }

  /*!
    \brief Constructs a masked polynomial matrix with a given mask.

    \throws ValueError if the size of the mask differs from the size of the matrix.
  */
  SDPMaskedPolyMatrix(const Matrix<Polynomial>& mat, const MatrixMask& mask)
    : Aref(mat)
  {
    if (mask.num_rows() != mat.num_rows() || mask.num_cols() != mat.num_cols())
      throw ValueError(ERROR_MSG("Matrix and mask of different sizes"));

    _mask.reset(new const_shared_ref<MatrixMask>(mask));
  }

  //! Returns a reference to the matrix.
  const Matrix<Polynomial>& matrix() const { return *Aref; }

  //! Returns a pointer to the mask, which is \c nullptr if there is no mask.
  const MatrixMask *mask() const
  { 
    if (_mask) return _mask->ptr();
    return nullptr;
  }

  //! Deleted assignment operator.
  SDPMaskedPolyMatrix& operator=(const SDPMaskedPolyMatrix& other) = delete;
};

/*!
  \brief A map associating polynomial matrices with variables for an SDP.

  The left-hand side of a polynomial constraint of the semidefinite
  programming problems that we consider  (see \ref tut_sdp for details)
  is an expression like

  \f[\sum_{j=1}^r \langle A_j, X_j \rangle,\f]

  where the \f$A_j\f$ are matrices of Laurent polynomials of the
  appropriate sizes.

  Class \c SDPPolynomialCombination represents expressions of this
  sort. It is a map associating instances of \c SDPMaskedPolyMatrix
  to variables of the problem.

  \section sdppol_iter Iteration

  Iteration is done over pairs <tt>std::pair<SDPVariable,
  SDPMaskedPolyMatrix></tt>. Only constant iteration is allowed. The
  pairs are iterated over in the order imposed by the variable
  indices.

  \section sdppol_op Constructing polynomial combinations

  Operators are defined so that instances of \c
  SDPPolynomialCombination can be constructed as instances of \c
  SDPLinearCombination; see documentation \ref sdplin_op "there".

  The same \ref sdplin_complex "observations" regarding the complexity
  of constructing an instance of \c SDPLinearCombination apply to this
  procedure of building an instance of \c SDPPolynomialCombination.
*/
class SDPPolynomialCombination {
  typedef std::map<SDPVariable, SDPMaskedPolyMatrix> SummandMap;

  SummandMap summands;

public:
  /* Constructors */

  //! Constructs an empty combination.
  SDPPolynomialCombination() { } 

  /*!
    \brief Constructs an \c SDPPolynomialCombination assigning matrix \a mm to variable \c X.

    \throws ValueError if the matrix is not square or its size differs from that of the variable.
  */
  SDPPolynomialCombination(const SDPVariable& X, const SDPMaskedPolyMatrix& mm);
  
  //! Constant iterator.
  typedef SummandMap::const_iterator iterator;

  //! Returns \c const iterator to the beginning.
  iterator begin() const { return summands.begin(); }

  //! Returns \c const "past-the-end" iterator.
  iterator end() const { return summands.end(); }

  /*!
    \brief In-place addition.

    Adds the polynomial combination \a other to this one. We assume that
    the variables occurring in this combination are distinct than those
    occuring in \a other. If this is not the case, an exception is thrown.

    \throws ValueError if a variable occurs in both this combination and
    \a other. If this exception is thrown then the result of the sum is
    unspecified.
  */
  SDPPolynomialCombination& operator+=(const SDPPolynomialCombination& other);

  friend SDPPolynomialCombination operator+(const SDPPolynomialCombination& a, const SDPPolynomialCombination& b);
};

//
// Prototypes
//

/*! \relates SDPPolynomialCombination
  \brief Multiplication of a masked matrix and a variable.

  Multiplying an instance of \c SDPMaskedPolyMatrix by an instance of \c
  SDPVariable yields an instance of \c SDPPolynomialCombination
  associating the given masked matrix with the given variable.

  Since \c SDPMaskedPolyMatrix has a constructor accepting a reference to
  <tt>Matrix<Polynomial></tt>, \cpp will promote any instance of
  <tt>Matrix<Polynomial></tt> multiplied by an instance of \c SDPVariable to an
  instance of \c SDPMaskedPolyMatrix in order to make the multiplication
  work. This means that one can also multiply instances of \c
  <tt>Matrix<Polynomial></tt> with instances of \c SDPVariable.

  \throws ValueError if the matrix is not square or if its size
  differs from that of the variable.
*/
extern SDPPolynomialCombination operator*(const SDPMaskedPolyMatrix& mat, const SDPVariable& X);

/*!
  \brief Adds two polynomial combinations.

  \throws ValueError if the combinations have a variable in common.
*/
extern SDPPolynomialCombination operator+(const SDPPolynomialCombination& a, const SDPPolynomialCombination& b);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
